/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/DB/PIDBTree.cpp,v $
 * $Date: 2001/04/07 13:30:18 $
 *
 Description:
\*____________________________________________________________________________*/
//$SourceTop:$

#include <assert.h>
#include <string.h>
#include <ctype.h>

#include "PIDBTree.h"
#include "AutoDel.h"
#include "StrToken.h"

/*____________________________________________________________________________*\
 *
 Description:
	Some compilers want dummy copy constructors.
	Here are null references to enable dummy copy constructors.
\*____________________________________________________________________________*/
static TypeFrame *__pFrame = 0;
static TypeFrame &__tFrame = *__pFrame;

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void *PIDBTree::AddString( void *pV )
{
	return DuplicateString( (const char *)pV );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void PIDBTree::DestructTree( void *pV )
{
	PI_DELETE( (_PIDB *)pV );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
#define INDENT  "|   "
void PIDBTree::RenderTree( ostream &os, const void *pV, int iLevel )
{
	PIDBTree *pDB = (PIDBTree *)pV;
	assert( pDB );
	int iLastType = pDB->iLastType;
	for(int i=0; i<=iLastType; i++)
		{
		const TypeFrame &tFrame = pDB->GetTypeFrame( i );
		PIString sTypeName( tFrame.pName );
		if ( sTypeName==PIString::Empty() )
			{
			pi_ltoa( i, 10, sTypeName );
			};
		_PIDBIterator *pIter = pDB->GetIterator( i, 0, 0 );
		for(; pIter->AtValidElement(); pIter->Next() )
			{
			os << endl;
			for(int j=iLevel; j; j--)
				{ os << INDENT; };
			const char *pKey;
			void *pValue = pIter->Current( &pKey );
			PIString sSubtypeName;
			if ( i==PIDBTYPE_USER )
				{ sSubtypeName = ((UserDBType *)pValue)-> GetTypeFrame().pName; };
			os << "<" << sTypeName << sSubtypeName << "|" << (pKey?pKey:"")
				<< ">";
			if ( tFrame.fnRender )
				{ (tFrame.fnRender)( os, pValue, iLevel+1 ); }
			else
				{ os << (void *)pValue; };
			};
		PI_DELETE( pIter );
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void PIDBTree::RenderString( ostream &os, const void *pV, int )
{
	if ( pV )
		{
		os << (const char *)pV;
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void PIDBTree::DestructString( void *pV )
{
	PI_DELETE( [] pV );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void PIDBTree::RenderOpaque( ostream &os, const void *pV, int )
{
	os << (void *)pV;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void PIDBTree::RenderUser( ostream &os, const void *pV, int iLevel )
{
	if ( pV )
		{
		UserDBType &tUser = *( (UserDBType *)pV );
		if ( tUser.GetTypeFrame().fnRender )
			{ (tUser.GetTypeFrame().fnRender)(os, tUser.GetValue(), iLevel); }
		else
			{ os << (void *)pV; };
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void PIDBTree::DestructUser( void *pV )
{
	PI_DELETE( (UserDBType *)pV );
}

/*____________________________________________________________________________*\
 *
 Global declarations:
	Type frames for the built-in types.
\*____________________________________________________________________________*/
TypeFrame PIDBTree::aTypeFrames[ PIDBTree::NUM_TYPES ] =
{
	{ 
		"DBTree",
		"Program database tree",
		0,
		PIDBTree::RenderTree,
		PIDBTree::DestructTree
	},
	{ 
		"String",
		"Program database string",
		PIDBTree::AddString,
		PIDBTree::RenderString,
		PIDBTree::DestructString
	},
	{
		"RFC822",
		"RFC822 compliant string",
		PIDBTree::AddString,
		PIDBTree::RenderString,
		PIDBTree::DestructString
	},
	{
		"Opaque",
		"Pointer to internal data structure",
		0,
		PIDBTree::RenderOpaque,
		0
	},
	{
		"User/",
		"User defined type with typeframe",
		0,
		PIDBTree::RenderUser,
		PIDBTree::DestructUser	
	}
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class KeyValuePair
{
private:
	const char *pKey;
	void *pValue;

public:
	KeyValuePair( const char *pTheKey, void *pTheValue )
		:	pKey(pTheKey), pValue(pTheValue)	{};
	inline const char *Key()					{ return pKey; };
	inline void *Value()						{ return pValue; };
	inline void Delete( void (* fnDelete )(void *) )
		{ if (fnDelete) { (fnDelete)(pValue); };  PI_DELETE( this ); }; 
	inline void *Detach()
		{ void *pTmp = pValue; pValue = 0; PI_DELETE( this ); return pTmp; };
	inline void Replace( void *pNew, void (* fnDelete )(void *) )
		{ if (fnDelete) { (fnDelete)(pValue); };  pValue = pNew; };

};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class PIDBTreePropagatesIterator : public _PIDBIterator
{
private:
	_PIDBIterator *pIterator;
	_PIDB *pParent;
	int iType;
	const char *pKey;
	int iFlags;

public:
	PIDBTreePropagatesIterator( _PIDB *pTheParent, int iTheType,
			const char *pTheKey, _PIDBIterator *pTheIterator, int iTheFlags )
		:	pIterator( pTheIterator ),
			pParent( pTheParent ),
			iType( iTheType ),
			pKey( pTheKey ),
			iFlags( iTheFlags )
		{
		};

	virtual ~PIDBTreePropagatesIterator()
		{
		PI_DELETE( pIterator );
		};

	virtual bool AtValidElement()
		{
		assert( pIterator );
		if ( (!pIterator->AtValidElement()) && pParent )
			{
			PI_DELETE( pIterator );
			pIterator = pParent->GetIterator( iType, pKey, iFlags );
			pParent = pParent->GetParent();
			return AtValidElement();
			};
		return pIterator->AtValidElement();
		};

	virtual void *Current( const char **ppKey )
		{
		assert( pIterator );
		return pIterator->Current( ppKey );
		};

	/*
		Remove the entry at the current index 
	*/
	virtual void RemoveCurrent()
		{
		assert( pIterator );
		if ( !pIterator ) { return; };
		pIterator->RemoveCurrent();
		};

	/*
		Detach and return the current element
	*/
	virtual void *DetachCurrent()
		{
		assert( pIterator );
		if ( !pIterator ) { return 0; };
		return pIterator->DetachCurrent();
		};

	/*
		Replace the current element with the specified one
	*/
	virtual void ReplaceCurrent( void *pNew )
		{
		assert( pIterator );
		if ( !pIterator ) { return; };
		pIterator->ReplaceCurrent( pNew );
		};

	/*
		Advance the contained iterator to the next element. If the
		contained iterator is exhausted and its an iterator for the
		current tree and this tree has a parent, then make the 
		contained interator be the iterator for the parent.
	*/
	virtual void Next()			
		{
		assert( pIterator );
		pIterator->Next();
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class PIDBTreeAnyKeyIterator : public _PIDBIterator
{
protected:
	DblList *aLists;
	int iNumLists;
	int iCurrentList;
	DblListIterator iIter;
	const TypeFrame &tTypeFrame;

	/*
		If the iterator is pointing at an empty list then initialise
		the iterator to point at the next list
	*/
	inline void AdvanceToValidKey()
		{
		if ( !iIter.BadIndex() )
			{ return; };
		for(; iCurrentList<iNumLists; iCurrentList++ )
			{
			iIter = aLists[iCurrentList];
			if ( !iIter.BadIndex() )
				{
				iCurrentList++;
				return;
				};
			};
		};

	/* --- forbid copy constructor --- */
	PIDBTreeAnyKeyIterator( const PIDBTreeAnyKeyIterator &iI )
	:	aLists( 0 ),
		iNumLists( 0 ),
		iCurrentList( 0 ),
		iIter( iI.aLists[0] ),
		tTypeFrame( __tFrame )
		{ assert( 0 ); };

public:
	PIDBTreeAnyKeyIterator( DblList *aTheLists, int iTheNumLists,
			const TypeFrame &tTheTypeFrame )
	:	aLists( aTheLists ),
		iNumLists( iTheNumLists ),
		iCurrentList( 0 ),
		iIter( aTheLists[0] ),
		tTypeFrame( tTheTypeFrame )
		{
		iCurrentList++;
		AdvanceToValidKey();
		};

	virtual bool AtValidElement()					{ return !iIter.BadIndex(); };

	virtual void *Current( const char **ppKey ) 	
		{
		if ( ppKey )
			{ *ppKey = ((KeyValuePair *)iIter.Current())->Key(); };
		return ((KeyValuePair *)iIter.Current())->Value();
		};

	/*
		Remove and destroy the entry at the current index 
	*/
	virtual void RemoveCurrent()
		{
		assert( !iIter.BadIndex() );
		if ( iIter.BadIndex() )	{ return; };
		((KeyValuePair *)aLists[iCurrentList].Detach(iIter))
			->Delete(tTypeFrame.fnDelete);
		};

	/*
		Detach and return the current element
	*/
	virtual void *DetachCurrent()
		{
		assert( !iIter.BadIndex() );
		if ( iIter.BadIndex() )	{ return 0; };
		return ((KeyValuePair *)aLists[iCurrentList].Detach(iIter))->Detach();
		};

	/*
		Replace the current element
	*/
	virtual void ReplaceCurrent( void *pNew )
		{
		assert( !iIter.BadIndex() );
		if ( iIter.BadIndex() )	{ return; };
		((KeyValuePair *)iIter.Current())-> Replace( pNew, 
			tTypeFrame.fnDelete );
		};

    virtual void Next()
		{
		iIter++;
		AdvanceToValidKey();
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	This Iterator traverses a PIDBTree structure and matches only against
	elements with the appropriate keys
\*____________________________________________________________________________*/
class PIDBTreeSpecificKeyIterator : public PIDBTreeAnyKeyIterator
{
private:
	const char *pKey;

	/*
		If the iterator is not currently pointing at a valid key then
		advance it to point to the next one
	*/
	inline void AdvanceToValidKey()
		{
		for(; !iIter.BadIndex() && 
			((KeyValuePair *)iIter.Current())->Key()!=pKey; iIter++ );
		};
	
public:
	PIDBTreeSpecificKeyIterator( DblList &tList, const char *pTheKey,
			const TypeFrame &tTheTypeFrame )
	:	PIDBTreeAnyKeyIterator( &tList, 1, tTheTypeFrame ),
		pKey( pTheKey )
		{
		AdvanceToValidKey();
		};

	virtual void Next()			
		{
		iIter++;
		AdvanceToValidKey();
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	public, static
 Description:
	Get the symbol table entry for the key and type. Type RFC822 is
	not case sensitive. All the other types are.
\*____________________________________________________________________________*/
const char *PIDBTree::GetFastKey( const char *pKey, int iType )
{
	return PISymbolTable_LookupOrInsert( 
			pKey,
			(iType == PIDBTYPE_RFC822 ) ? 0 : 1 );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	public, static
 Description:
	Returns -1 on error, else the PIDBTYPE number for name pType
\*____________________________________________________________________________*/
int PIDBTree::PIDBTypeNameToNumber( const char *pType )
{
	assert( pType );
	if ( !pType )
		{ return PIDBTYPE_INVALID; };

	/* --- skip leading whitespace --- */
	for( ; *pType && (isspace(*pType)); pType++ );

	/* --- include only to first whitespace trailing whitespace --- */
	int j=0;
	for( ; pType[j] && !(isspace(pType[j])); j++ );
	/* --- j now contains the length of the first word in pType --- */

	/* --- the type --- */
	if ( !PIUtil_strncmpi( pType, PIDBTYPE_TEXT_TREE, j ) )
		{ return PIDBTYPE_TREE; }
	else if ( !PIUtil_strncmpi( pType, PIDBTYPE_TEXT_STRING, j ) )
		{ return PIDBTYPE_STRING; }
	else if ( !PIUtil_strncmpi( pType, PIDBTYPE_TEXT_RFC822, j ) )
		{ return PIDBTYPE_RFC822; }
	else if ( !PIUtil_strncmpi( pType, PIDBTYPE_TEXT_OPAQUE, j ) )
		{ return PIDBTYPE_OPAQUE; }
	else if ( !PIUtil_strncmpi( pType, PIDBTYPE_TEXT_USER, j ) )
		{ return PIDBTYPE_USER; }
	else
		{
		};
	return PIDBTYPE_INVALID;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	public, static
 Description:
	Return numeric flags mask given the bar delimited list pFlags.
\*____________________________________________________________________________*/
int PIDBTree::PIDBFlagsNameToNumber( const char *pFlags )
{
	int iFlags = PIDBFLAG_NONE;
	if ( pFlags )
		{
		StringTokenizer tTokens( pFlags, "|" );
		for(int i=0; i<tTokens.NumTokens(); i++)
			{
			const char *pToken = tTokens.GetToken( i );

			/* --- skip leading whitespace --- */
			for( ; *pToken && (isspace(*pToken)); pToken++ );

			/* --- include only to first whitespace trailing whitespace --- */
			int j=0;
			for( ; pToken[j] && !(isspace(pToken[j])); j++ );
			/* --- j now contains the length of the first word in pToken --- */

			if ( !PIUtil_strncmpi( pToken, PIDBFLAG_TEXT_PROPAGATEUP, j ) )
				{ iFlags |= PIDBFLAG_PROPAGATEUP; }
			else if ( !PIUtil_strncmpi( pToken, PIDBFLAG_TEXT_NOCALLBACK, j ) )
				{ iFlags |= PIDBFLAG_NOCALLBACK; }
			else if ( !PIUtil_strncmpi( pToken, PIDBFLAG_TEXT_FASTKEY, j ) )
				{ iFlags |= PIDBFLAG_FASTKEY; }
			else
				{ assert( 0 ); /* --- catch this at development time --- */ };
			};
		};
	return iFlags;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int PIDBTree::Add( int iType, const char *pKey, void *pValue,
	int iFlags )
{
	assert( iType<=iLastType );
	if ( iType>iLastType )
		{ return PIAPI_EINVAL; };

	/* --- get the standard address for this string --- */
	const char *pSym = 
		( iFlags & PIDBFLAG_FASTKEY ) ? 
			pKey : 
			( pKey ? PIDBTree::GetFastKey( pKey, iType ) : 0 );
	const TypeFrame &tTypeFrame = GetTypeFrame( iType );
	void *pNewValue = tTypeFrame.fnAdd ? (tTypeFrame.fnAdd)(pValue) : pValue;
	GetList( pTypes, iType, pSym ).Append( (DblList::type)
		PI_NEW( KeyValuePair( pSym, pNewValue ) ) );
	if ( iType==PIDBTYPE_TREE && pValue )
		{
		((_PIDB *)pValue)->SetParent( this );
		};
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int PIDBTree::Replace( int iType, const char *pKey, void *pValue,
	int iFlags )
{
	_PIDBIterator *pIter;
	AutoDelete<_PIDBIterator> tA( pIter = GetIterator( iType, pKey, iFlags ) );
	(void)tA;
	if ( pIter )
		{
		if ( pIter->AtValidElement() )
			{
			const TypeFrame &tTypeFrame = GetTypeFrame( iType );
			void *pNewValue = tTypeFrame.fnAdd ? (tTypeFrame.fnAdd)(pValue) :
					pValue;
			pIter->ReplaceCurrent( pNewValue );
			return PIAPI_COMPLETED;
			};
		};
	return Add( iType, pKey, pValue, iFlags );  
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
_PIDBIterator *PIDBTree::GetIterator( int iType, const char *pKey, int iFlags )
{
	if ( iType<0 || iType>iLastType ) return 0;
	const TypeFrame &tTypeFrame= GetTypeFrame( iType );
	const char *pSym = 
		( iFlags & PIDBFLAG_FASTKEY ) ? 
			pKey : 
			( pKey ? PIDBTree::GetFastKey( pKey, iType ) : 0 );
	_PIDBIterator *pIter = 0;
	pIter = pSym ? 
		PI_NEW( PIDBTreeSpecificKeyIterator(
				GetList( pTypes, iType, pSym ), pSym, tTypeFrame ) ):
		PI_NEW( PIDBTreeAnyKeyIterator( 
				&( pTypes[GetFirstListIndexForType(iType)] ),
				HASH_TABLE_SIZE, tTypeFrame  ) );
	if ( (iFlags & PIDBFLAG_PROPAGATEUP ) && GetParent() ) 
		{
		return PI_NEW( PIDBTreePropagatesIterator( GetParent(), iType,
			pSym, pIter, iFlags ) );
		}
	else
		{
		return pIter;
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void *PIDBTree::Lookup( int iType, const char *pKey, int iFlags )
{
#if 1	/* change this to 0 to disable this optimization */
	/* 
	** Optimization:
	** First check if the PROPAGATEUP flag is not set, if not then
	** do the lookup here without generating an iterator.
	*/
	if ( !(iFlags & PIDBFLAG_PROPAGATEUP) )
		{
		/* --- get fast key --- */
		const char *pSym = 
			( iFlags & PIDBFLAG_FASTKEY ) ? 
				pKey : 
				( pKey ? PIDBTree::GetFastKey( pKey, iType ) : 0 );

		const char *pTmp = pSym;
		for( DblListIterator i( GetList( pTypes, iType, pTmp ) );
			!i.BadIndex(); i++ )
			{
			if ( ((KeyValuePair *)i.Current())->Key()==pSym )
				{
				return ((KeyValuePair *)i.Current())->Value();
				};
			};

		/* --- not found --- */
		return 0;
		};
#endif

	/* ---
	resort to older slower lookup mechanism
	--- */
	_PIDBIterator *pIter;
	AutoDelete<_PIDBIterator> tA( pIter = GetIterator( iType, pKey, iFlags ) );
	(void)tA;
	if ( !pIter )
		{ return 0; };
	if ( pIter->AtValidElement() )
		{
		void *pV = pIter->Current( 0 );
		if ( !pV )
			{
			return 0;
			};
		return pV;
		}
	else
		{
		return 0;
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PIDBTree::PIDBTree(	_PIDB *pTheParent, const char *pName,
				int iNumAdditionalTypes,
				TypeFrame *aTheAdditionalTypeFrames )
:	_PIDB( pTheParent ),
	iLastType( iNumAdditionalTypes + PIDBTYPE_LAST ),
	aAdditionalTypeFrames( aTheAdditionalTypeFrames )
{
	if ( pTheParent )
		{
		pTheParent->Add( PIDBTYPE_TREE, pName, this, 0 );
		}; 
	assert( iLastType==PIDBTYPE_LAST || aAdditionalTypeFrames );
	pTypes = PI_NEW( DblList[ GetNumLists( iLastType ) ] );	
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PIDBTree::~PIDBTree()
{
	/* --- destruct the builtin and additional types --- */
	int j=0;
	int i;
	for(i=0; j<PIDBTYPE_BUILTINSIZE; i++, j++)
		{
		InternalDestruct(
			&( pTypes[ GetFirstListIndexForType( j ) ] ),
			HASH_TABLE_SIZE,
			aTypeFrames[i].fnDelete
			);
		};

	assert( j==PIDBTYPE_BUILTINSIZE );
	for(i=0; j<=iLastType; i++, j++)
		{
		InternalDestruct(
			&( pTypes[ GetFirstListIndexForType( j ) ] ),
			HASH_TABLE_SIZE,
			aAdditionalTypeFrames[i].fnDelete );
		};

	PI_DELETE( [] pTypes );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void PIDBTree::InternalDestruct(
		DblList *paLists, int iNumLists, void (* fnDelete)(void *) )
{
	for( int j=0; j<iNumLists; j++ )
		{
		DblList &tList = paLists[j];
		for(DblListIterator i(tList); !i.BadIndex(); i++)
			{
			((KeyValuePair *)i.Current())->Delete( fnDelete );
			};
		};
}

