/*____________________________________________________________________________*\

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Pi3API/_Pi3Expr.cpp,v $
 * $Date: 2002/06/04 15:50:16 $
 *
 Description:

	Description:
	------------
	Expressions are snippets of a mini-lanaguage for generating and
	manipulating text. In general it a pretty simple concept, - expressions
	can be simple expressions like static strings, or can be more complex
	like list of different types of expressions (ExpressionComponent) or
	expression whose results are a function of thier subexpressions 
	(FunctionComponents).

	The only other complicating factor is that expressions try to 
	optimize themselves by merging and preevaluating static expressions
	at the construction phase.

	There are three kinds of classes used internally here:-

		- Context, this contains the context for an expression evaluation
		including a PIHTTP object (which may be empty if this is an
		optimization (compilation) of a static expression). This class
		also contains parameter information set by the expression 
		evaluator.

		- Components, all derived classes of ComponentBase. They form
		of the core of expressions and are polymorphic expressions of
		various types.

		- Pi3Expression, this is a wrapper around a single 
		ExpressionComponent and is the interface to the outside world.

	NOTES:
	------
	Make this more efficient by
		- evaluate all shortcuts and update short documentation 
		- maybe change the way that Context object are generated for the
		compile phase, they should probably be passed in as arguments. This
		also supports run-time compilation which would resolve everything
		to a string.
		- A little genernal cleanup is in order here. In particular,
		ditch FunctionArgument classes in favour of using 
		Components directly. And get rid of the automatic creation of empty
		PIHTTP and Context classes during optimization, instead make one
		single one and pass it around.
		- bug, it is an error that the following

			(Process, thread) ($P, $k)

			evaluates to

			Process/thread -1/508

\*____________________________________________________________________________*/
//$SourceTop:$

#include <stdio.h>
#include <ctype.h>

#include "PIStrStr.h"
#include "_Pi3Expr.h"
#include "AutoDel.h"
#include "HTTPCore.h"
#include "HTTPUtil.h"

#define INDENT_PATTERN "\t"
/*
#define D { cerr << __FILE__ << ", " << __LINE__ << endl; };
*/
#define D

/*
** NOTE: This is because of some compiler problems
*/
void *_Pi3Expression::Compiler_Bug_Alloc( void *p1, void *p2,
	void *p3 )
	{
	return PI_NEW( _Pi3Expression( (const char *)p1, (FnPi3Write *)p2,
			*( (PIString *)p3 ) ) );
	}

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Documentation.
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation 
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	Pi3Expression

Description:
	A Pi3Expression is a general purpose expression syntax for efficient
	dynamic creation of text based on varied criteria. The expression takes
	the form of a sequence of text characters with some control characters
	which modify the meaning of the text which follows them. A Pi3Expression
	does not have to be enclosed in quotes but often will be quoted in
	configuration files to deliniate the expression from other configuration
	components.
	For illustration all examples used here will enclose the example expression
	in double-quotations ("").

Components:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Component
<TH>Syntax
<TH>Short Description
<TH>Example Expression
<TH>Example Result

<TR>
<TD>String
<TD>Any character except $,%,&amp;
<TD>Plain text
<TD>"Hello, World!"
<TD>Hello, World!

<TR>
<TD>Escape Sequences
<TD>\c, where c is any character
<TD>Escaped charcaters, like printf format
<TD>"Line1\nLine2"
<TD>Line1<BR>Line2

<TR>
<TD>Literal
<TD>'any characters'
<TD>Literal text, characters not interpreted.
<TD>"The percent character is '%'"
<TD>The percent character is %

<TR>
<TD>Parameter
<TD>%c, where c is a letter
<TD>A parameter, value depends on Pi3Expression context
<TD>"Parameter value of parameter ''v'' is ''%v''"
<TD>Parameter value of parameter 'v' is 'john'

<TR>
<TD>Shortcut
<TD>$s, where s is a letter
<TD>A shortcut, shortcut value mappings are given below
<TD>"The current process id is $P"
<TD>The current process id is 24456

<TR>
<TD>Functions
<TD>&amp;FunctionName(Param1,Param2,..,ParamN)
<TD>A text processing function, described in 'Pi3Expression_Functions'
<TD>"Process id truncated to two characters is &amp;trunc($P,2)"
<TD>Process id truncated to two characters is 24

</TABLE>

<H4>Detailed Description of Components</H4>
<H5>
	String
</H5>
Regular text that does not conform to other expression components is
reproduced exactly if the expression result as in the expression string.
The characters '(', ',', and ')' are considered normal characters outside
a function context.

<H5>
	Escape Sequences
</H5>
The escape or backslash character will be expanded to a control character in accordance with the following table:-
<CENTER>
<TABLE BORDER=1>
<TH>Escape Sequence
<TH>Meaning

<TR>
<TD>\a
<TD>Alert

<TR>
<TD>\b
<TD>Backslash

<TR>
<TD>\f
<TD>Form-feed

<TR>
<TD>\n
<TD>New-line

<TR>
<TD>\r
<TD>Carriage return

<TR>
<TD>\t
<TD>Horizontal tab

<TR>
<TD>\v
<TD>Vertical tab
</TABLE>
</CENTER>
Additionally any other character including characters with special meanings
in Pi3Expressions ($,%,&amp;,\,&acute;) can be specified by placing &acute;\&acute; before them.

<H5>
	Literal
</H5>
Character sequences enclosed in single quotes will be reproduced in
the expression result as in the expression specification, but with the
single quotes ommitted. 

<H5>
	Parameter
</H5>
The '%' character followed by a letter specifies a parameter. The
value of a parameter depends on the evaluation context and should be
documented along with the facility that uses the expression. Parameter
characters are case sensitive.

<H5>
	Shortcut
</H5>
The '$' character followed by a letter specifies a shortcut. Shortcut
characters are case sensitive. The follow table gives shortcut mappings.
<CENTER>
<TABLE BORDER=1>
<TH>Shortcut<TH>Evaluates to<TR>
<TD>$a<TD>E-mail address of administrator (for current virtual host)<TR>
<TD>$A<TD>IP Address of remote client<TR>
<TD>$b<TD>Decimal representation of number of bytes sent to client this request<TR>
<TD>$c<TD>Content-Type for the response<TR>
<TD>$C<TD>Content-Type for the request (incoming)<TR>
<TD>$d<TD>Debug flag, returns non-zero length string if debugging is on<TR>
<TD>$D<TD>Handler time delta. Time spent in last handler object<TR>
<TD>$E<TD>All error messages logged while handling this (sub)request<TR>
<TD>$f<TD>Physical path to resource<TR>
<TD>$g<TD>Indicate whether or not SSL is being used (HTTPS), (on/off)<TR>
<TD>$G<TD>SSL secret key size<TR>
<TD>$h<TD>Hostname of client<TR>
<TD>$H<TD>Request HTTP protocol stamp<TR>
<TD>$i<TD>Remoteident of client<TR>
<TD>$I<TD>Path info section of client request URL<TR>
<TD>$k<TD>Thread id of executing thread of execution<TR>
<TD>$K<TD>SSL public key size<TR>
<TD>$l<TD>Local IP address of connection<TR>
<TD>$m<TD>Request method from client<TR>
<TD>$M<TD>End of line sequence for the current operating system<TR>
<TD>$n<TD>Name of current handler object<TR>
<TD>$N<TD>Name of current handler phase<TR>
<TD>$o<TD>Objectname of virtual host<TR>
<TD>$p<TD>Port associated with this virtual host<TR>
<TD>$P<TD>Process id<TR>
<TD>$q<TD>Client query string<TR>
<TD>$r<TD>Full uninterpreted request line from client<TR>
<TD>$R<TD>Name of result code from last handler<TR>
<TD>$s<TD>Current decimal HTTP response status code<TR>
<TD>$S<TD>Server version stamp<TR>
<TD>$t<TD>Time in common logfile format<TR>
<TD>$T<TD>Elapsed since request context (PIHTTP) was created<TR>
<TD>$u<TD>Remote username (from authentication)<TR>
<TD>$U<TD>URL path of original request<TR>
<TD>$v<TD>Hostname of virtual host<TR>
<TD>$x<TD>Authentication type from is access was authenticated<TR>
<TD>$X<TD>Depth of current request in subrequest hierarchy<TR>
<TD>$y<TD>Content-Length: outgoing to client<TR>
<TD>$Y<TD>Content-Length: incoming from client<TR>
<TD>$z<TD>Script name<TR>
<TD>$Z<TD>Path translated (url-unencoded path info)<TR>
</TABLE>
</CENTER>

<H5>
	Functions
</H5>
The built-in text manipulation functions are described in the section
'Pi3Expression_Functions'.

Examples:
<CENTER>
<TABLE BORDER=1>
	<TD><B>Expression</B><TD>Process/Thread information [$P|$k]<TR>
	<TD><B>Evaluates to</B><TD>Process/Thread information [24965|23]<TR>
</TABLE>
<TABLE BORDER=1>
	<TD><B>Expression</B><TD>[$t]$h|$r<TR>
	<TD><B>Evaluates to</B><TD>|15/May/1997:21:30:16 +0000|caught.evil.dom|GET /../../etc/passwd HTTP/1.0<TR>
</TABLE>
</CENTER>

/*___+++HTMLDOC_END+++___*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	Pi3Expression_Functions

Description:
	Pi3Expressions can use built in functions for text manipulation and
	conditional processing. These functions can be nested to perform
	complex evaluations. The built-in text processing functions are
	described here.

	Where text is interpreted as a numeric value, leading whitespace is
	skipped, then the text before the first non-digit is interpreted as
	a decimal number.

	Where text is interpreted as a boolean value, any non-zero length is
	considered 'true' and any text of length zero is considered 'false'.

Functions:
<H5>Overview</H5>
	Note the following conventions used to describing these functions:-
<LIST>
<ITEM>text - any text, e.g. "foo", "234'&amp;'ad"
<ITEM>number - text interpreted as a number, e.g. "5a" (interpreted as 5).
<ITEM>[field] - an optional argument
<ITEM>value1|value2 - either "value1" or "value2" as text, e.g left|right
</LIST>

<TABLE BORDER=1>
<TH>Function
<TH>Number of Arguments
<TH>Syntax
<TH>Short Description
<TH>Example Expression
<TH>Example Result

<TR>
<TD>abbrev
<TD>3
<TD>&amp;abbrev(text,number,ellipse)
<TD>truncate text and append other text if truncated
<TD>"&amp;abbrev(Abcdef,7,...)"<BR>"&amp;abbrev(Abcdefgh,7,...)"
<TD>"Abcdef"<BR>"Abcd..."

<TR>
<TD>align
<TD>2..4
<TD>&amp;align(text,number,[left|right],[padding])
<TD>Align text within a field
<TD>"&amp;align(1234,7)"<BR>"&amp;align(1234,7,right,-)"
<TD>"1234&nbsp;&nbsp;&nbsp;"<BR>"---1234"

<TR>
<TD>arg,_
<TD>0..2
<TD>&amp;_([index],number)
<TD>Function argument
<TD>"&amp;_(0,0)"; &amp;arg(1)
<TD>context dependent

<TR>
<TD>cmp
<TD>2
<TD>&amp;cmp(text1,text2)
<TD>String comparision
<TD>"&amp;cmp(hello,Hello)"
<TD>""

<TR>
<TD>cmpi
<TD>2
<TD>&amp;cmpi(text1,text2)
<TD>String comparision case insensitive
<TD>"&amp;cmp(hello,Hello)"
<TD>"true"
<TR>

<TR>
<TD>dblookup
<TD>3..7
<TD>&amp;dblookup(dbname,type,variable[,flags][,fnvalue][,first][,last])
<TD>Lookup a value in an internal database
<TD>"&amp;dblookup(request,rfc822,Content-Type)"
<TD>"application/x-www-form-urlencoded"

<TR>
<TD>dbreplace
<TD>3..7
<TD>&amp;dbreplace(dbname,type,variable,value[,flags])
<TD>Replace a value in an internal database
<TD>"&amp;dbreplace(request,rfc822,Content-Type,text/html)"
<TD>"text/html"

<TR>
<TD>if
<TD>2..3
<TD>&amp;if(boolean,true_text,false_text)
<TD>Conditionally include text
<TD>"&amp;if(text,yes,no)"<BR>"&amp;if(,yes,no)"
<TD>"yes"<BR>"no"

<TR>
<TD>not
<TD>1
<TD>&amp;not(boolean)
<TD>Negate boolean argument
<TD>"&amp;not(text)"<BR>"&amp;not()"
<TD>""<BR>"true"

<TR>
<TD>regexp
<TD>2
<TD>&amp;regexp(regexp,text)
<TD>Regular expression comparision
<TD>"&amp;regexp(j*n,john)"<BR>"&amp;regexp(b*l,john)"
<TD>"true"<BR>""

<TR>
<TD>trunc
<TD>2
<TD>&amp;trunc(text,number)
<TD>Truncate text to length of number
<TD>"&amp;trunc(0123,2)"
<TD>"01"

</TABLE>

<H4>Detailed Description of Functions</H4>
<H5>
	abbrev(field,fieldlen,append)
</H5>
This function is used to truncate text to a field of fieldlen characters
appending text 'append' at the end of the field if it was truncated, this
is useful for truncating test to fit in a field while given a clear
indication if the field was concatenated.

<H5>
	align(text,length[,left|right][,padding])
</H5>
Write text 'text' into a field of length characters, aligning to the
left or right. The character sequence padding is repeated to
fill out the field.
The default alignment is left. The default padding is spaces.
If 'text' is longer than the value of 'length', it will not be truncated.
A field of exactly 'length' character can be ensured by using an expression like:-
<CODE>
&amp;align(&amp;trunc(text is too big,10),10)
</CODE>

<H5>
	arg([index],[context])
</H5>
This is a special function with meaning only within the context of another
function which is iterating through a number of elements. The argument
'index' is the index of the argument from the enclosing function. The
argument context is a number indicating the which enclosing functions
context to use. Context 0 refers to the callees context, 1 to the callees 
callee etc.

<H5>
	cmp(text1,text2), cmpi(text1,text2)
</H5>
Respectively case sensitive and insensitive comparision of two text
patterns.

<H5>
	dblookup(dbname,type,variable[,flags][,fnvalue][,first][,last])
</H5>
<STRONG>NOTE:</STRONG>This function is not currently supported to the full
specification as described here.
Iterate over a number of elements of a specific type in a particular database.
<CENTER>
Function arguments:-
<TABLE BORDER=1>
<TH>Argument position
<TH>Description
<TH>Default, (+)=Mandatory
<TH>Values

<TR>
<TD>1
<TD>name of database
<TD>+
<TD>program|connection|host|request|response

<TR>
<TD>2
<TD>type of element
<TD>+
<TD>rfc822|string|opaque

<TR>
<TD>3
<TD>variable name
<TD>+
<TD>a variable name (or empty for all)

<TR>
<TD>4
<TD>db flags affecting lookup
<TD>0
<TD>pidbflag_propagateup|pidbflag_fastkey|0

<TR>
<TD>5
<TD>callback argument
<TD>&amp;_()
<TD>Expression invoked for value of every db entry 

<TR>
<TD>6
<TD>start index
<TD>0
<TD>Number

<TR>
<TD>7
<TD>end index
<TD>0
<TD>Number

</TABLE>
</CENTER>


<H5>
	dbreplace(dbname,type,variable,value[,flags])
</H5>
Replace the value of a variable of the specific type in a particular
database. The function arguments are the same as described above.

<H5>
	if(boolean,true_text,false_text)
</H5>
If boolean is true (i.e., boolean is any character sequence with length
greater than zero), then this function evaluates to true_text, otherwise this
function evaluates to false_text.

<H5>
	not(boolean)
</H5>
If boolean is true (i.e., boolean is any character sequence with length
greater than zero), then this function returns a zero length text string
(false), otherwise a non-zero length text string (true) is returned.
The net result is the negation of the boolean argument.

<H5>
	regexp(regexp,text)
</H5>
Determine if the regular expression 'regex' matches the text 'text'. 
Returns 'true' (non-empty text) on success, 'false' (empty text) on
failure.
Regular expression matching is done using the function HTTPUtil_regexMatch(),
reference that function to see which regular expression constructs which
are supported.

<H5>
	trunc(text,number)
</H5>
Truncate text to the the decimal value of number.

/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Component context, largely used for nested functions.
\*____________________________________________________________________________*/
class Context
{
public:
	PIHTTP &tPIHTTP;
	enum { CONTEXT_SIZE=10 };
	Context *pParent;
	void *pData;
	const char *aString[CONTEXT_SIZE];
	int aLength[CONTEXT_SIZE];

	Context(
			Context *pTheParent,
			PIHTTP &tThePIHTTP,
			void *pTheData )
	:	tPIHTTP( tThePIHTTP ),
		pParent( pTheParent ),
		pData( pTheData )
		{
		memset( aString, 0, sizeof( const char * ) * CONTEXT_SIZE );
		memset( aLength, 0, sizeof( int ) * CONTEXT_SIZE );
		};

	/* --- traverse up list of contexts --- */
	inline Context *GetContext( int iOffset )
		{
		assert( iOffset>=0 );
		if ( !pParent )
			{ return 0; };
		if ( iOffset==0 )
			{ return this; };
		return  pParent->GetContext( iOffset-1 );
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class StringComponent : public ComponentBase
{
private:
	PIString sString;

public:
	/* --- string with length iLen --- */
	StringComponent( const char *pString, int iLen )
	:	sString( pString, iLen )
		{};

	/* --- null terminated string --- */
	StringComponent( const char *pString )	
	:	sString( pString )
		{};

	virtual void WriteDebugTree( int iIndent ) const
		{
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "String:'" << sString << "'" << endl;
		};

	virtual int Write( Context &, char *pszBuffer, int iLength ) const
		{
		if ( iLength>0 )
			{
			assert( pszBuffer );
			strncpy( pszBuffer, sString, iLength );
			};
		return sString.Len();
		};

	virtual int IsOK() const		{ return 1; };
	virtual int IsStatic() const	{ return 1; };
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class ExpressionComponent : public ComponentBase
{
private:
	DblList lComponents;
	int iOK;
	int iStatic;

	/* --- internal methods --- */
	int Compile( Context &tContext, DblList &lRawComponents,
		PIString & /* sError */ )
		{
		/* ---
		optimize the expression by merging static components 
		together
		--- */

		/* --- make a really big buffer --- */
		enum { BUF_SIZE=32767 };
		char *pszBuffer = PI_NEW( char[BUF_SIZE+1] );	/* too big for auto */
		assert( pszBuffer );
	
		/* --- iterate over components merging where possible --- */
		int iLength = 0;	/* --- length of buffer with static data --- */
		iStatic = 1;		/* --- mark this static till we know otherwise ---*/
		for( ; lRawComponents.Size() ; )
			{
			ComponentBase *pComp = (ComponentBase *)lRawComponents.Cap();
			assert( pComp );
			if ( pComp->IsStatic() )
				{
				/* --- static --- */
				int iWritten = pComp->Write( tContext, &(pszBuffer[iLength]),
					BUF_SIZE-iLength );
				assert( iWritten!=-1 );
				if ( iWritten==-1 )
					{
					PI_DELETE( pComp );
					PI_DELETE( [] pszBuffer );
					return 0;
					};
				if ( (iLength+iWritten)<=BUF_SIZE )
					{
					/* --- component is written into buffer, start next one --*/
					PI_DELETE( pComp );
					iLength+=iWritten;
					continue;
					};
				/* --- else don't merge this component --- */
				};
				
			/* --- flush the buffer --- */
			if ( iLength )
				{
				/* --- previous components have been written --- */
				lComponents.Append( (DblList::type)PI_NEW(
					StringComponent( pszBuffer, iLength ) ) );
				iLength = 0;
				};
			iStatic = 0;
			lComponents.Append( (DblList::type)pComp );
			};

		if ( iLength )
			{
			/* --- previous components have been written --- */
			lComponents.Append( (DblList::type)PI_NEW(
				StringComponent( pszBuffer, iLength ) ) );
			iLength = 0;
			};
	
		PI_DELETE( [] pszBuffer );
		return 1;
		};

	/* --- forbid copy constructor --- */
	ExpressionComponent( const ExpressionComponent & )
		{ assert( 0 ); };

public:
	ExpressionComponent( Context &tContext, DblList &lRawComponents,
		PIString &sError )
	:	iOK( 0 ),
		iStatic( 0 )
		{
		if ( Compile( tContext, lRawComponents, sError ) )
			{ iOK = 1; };
		};

	~ExpressionComponent()
		{
		for( DblListIterator i( lComponents ); !i.BadIndex(); i++ )
			{
			ComponentBase *pComp = (ComponentBase *)i.Current();
			PI_DELETE( pComp );
			};
		};

	virtual void WriteDebugTree( int /* iIndent */ ) const
		{
#if 0
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "String:'" << sString << "'" << endl;
#endif
		};

	virtual int Write( Context &tContext, char *pszBuffer, int iMaxLen ) const
		{
		int iWrittenSoFar = 0;
		int iLengthSoFar = 0;
		for( ConstDblListIterator i( lComponents ); !i.BadIndex(); i++ )
			{
			ComponentBase *pComp = (ComponentBase *)i.Current();
			assert( pComp );
			int iWritten = pComp->Write( tContext,
					&( pszBuffer[iWrittenSoFar] ), iMaxLen-iWrittenSoFar );

			if ( iWritten==-1 )
				{
				assert( 0 );
				return -1;
				};

			/* else */
			iLengthSoFar += iWritten;

			if ( iLengthSoFar>iMaxLen )
				{
				iWritten = iMaxLen - iWrittenSoFar;
				};

			iWrittenSoFar += iWritten;
			assert( iWrittenSoFar<=iMaxLen );
			};

		return iLengthSoFar;
		};

	virtual int IsOK() const
		{
		return iOK;
		};
	virtual int IsStatic() const	
		{
		return iStatic;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class ParameterComponent : public ComponentBase
{
private:
	FnPi3Write tFunction;

public:
	ParameterComponent( FnPi3Write tTheFunction )	
	:	tFunction( tTheFunction )
		{
		if ( !tTheFunction )
			{ return; };
		}

	virtual void WriteDebugTree( int iIndent ) const 
		{
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "Parameter:" << endl;
		};

	virtual int Write( Context &tContext, char *pszBuffer, int iLength ) const
		{
		assert( tFunction );
		return (tFunction)(&(tContext.tPIHTTP), tContext.pData, pszBuffer,
			iLength );
		};

	virtual int IsOK() const { return tFunction!=0; };
	virtual int IsStatic()	const { return 0; /* some params only */ };
};

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
typedef int (* SCFnWrite)( Context &, char *, int );
int iShortcutsInitialized = 0;
static SCFnWrite aFunctions[256];

/* --- macros to simplify shortcut functions --- */
#define O PIDBTYPE_OPAQUE
#define S PIDBTYPE_STRING
#define F PIDBTYPE_RFC822
#define C(x,y) PIDB_lookup( tContext.tPIHTTP.GetConnectionDB(), x, y, 0 )
#define Q(x,y) PIDB_lookup( tContext.tPIHTTP.GetRequestDB(), x, y, 0 )
#define R(x,y) PIDB_lookup( tContext.tPIHTTP.GetResponseDB(), x, y, 0 )
#define V(x,y) PIDB_lookup( tContext.tPIHTTP.GetHostDB(), x, y, 0 )
#define SC_STR_FUNCTION(fn, value)\
static int (fn)( Context &tContext, char *pszBuffer, int iLength ) \
{ \
	(void)tContext; \
	const char *pString = (const char *)(value); \
	if ( !pString )	{ return 0; }; \
	int iValueLen = strlen( pString ); \
	int iLen = iValueLen < iLength ? iValueLen : iLength;	\
	if ( iLen>0 ) \
		{ \
		assert( pszBuffer ); \
		strncpy( pszBuffer, pString, iLen ); \
		}; \
	return iValueLen; \
}
#	define SC_NUM_FUNCTION(fn, value, flags)\
static int (fn)( Context &tContext, char *pszBuffer, int iLength ) \
{ \
	(void)tContext; \
	enum { BUF_SIZE=63 }; \
	char szBuf[BUF_SIZE+1]; \
	*szBuf = '\0'; \
	sprintf( szBuf, flags, (value) ); \
	int iValueLen = strlen( szBuf ); \
	int iLen = iValueLen < iLength ? iValueLen : iLength;	\
	if ( iLen>0 ) \
		{ \
		assert( pszBuffer ); \
		strncpy( pszBuffer, szBuf, iLen ); \
		}; \
	return iValueLen; \
}

SC_STR_FUNCTION( sc_a, V( F, "Administrator") );
SC_STR_FUNCTION( sc_A, C( S, KEY_INT_REMOTEADDR ) );
SC_NUM_FUNCTION( sc_b, tContext.tPIHTTP.GetIOBuffer().GetBytesSent(), "%d" );
SC_STR_FUNCTION( sc_c, R( F, KEY_HTTP_CONTENTTYPE ) );
SC_STR_FUNCTION( sc_C, Q( F, KEY_HTTP_CONTENTTYPE ) );
SC_STR_FUNCTION( sc_d, HTTPCore_debugEnabled() ? "true" : "" );
SC_NUM_FUNCTION( sc_D, tContext.tPIHTTP.iDelta, "%d" );
SC_STR_FUNCTION( sc_f, R( S, KEY_INT_PATH ) );
SC_STR_FUNCTION( sc_G, C( S, KEY_HTTPS_SECRETKEYSIZE ) );
SC_STR_FUNCTION( sc_h, C( S, KEY_INT_REMOTEHOST ) );
SC_STR_FUNCTION( sc_H, Q( S, KEY_HTTP_PROTOCOL ) );
SC_STR_FUNCTION( sc_i, C( S, KEY_INT_REMOTEIDENT ) );
SC_STR_FUNCTION( sc_I, R( S, KEY_INT_PATHINFO ) );
SC_NUM_FUNCTION( sc_k, (long)PIThread_getSystemHandle( PIThread_getCurrent() ),
		"%ld");
SC_STR_FUNCTION( sc_K, C( S, KEY_HTTPS_KEYSIZE ) );
SC_STR_FUNCTION( sc_m, Q( S, KEY_HTTP_METHOD ) );
#if POSIX
SC_STR_FUNCTION( sc_M, "\n" );
#else
SC_STR_FUNCTION( sc_M, "\r\n" );
#endif
SC_STR_FUNCTION( sc_n, tContext.tPIHTTP.pHandlerName );
SC_STR_FUNCTION( sc_N, HTTPUtil_phaseNumberToName(
		tContext.tPIHTTP.ciPhase ) );
SC_STR_FUNCTION( sc_l, C( S, KEY_INT_LOCALADDR ) );
SC_STR_FUNCTION( sc_o, V( F, KEY_INT_OBJECTNAME ) );
SC_STR_FUNCTION( sc_p, HTTPUtil_getHostPort(&(tContext.tPIHTTP)) );
SC_NUM_FUNCTION( sc_P, PIPlatform_getProcessId(), "%d" );
SC_STR_FUNCTION( sc_q, Q( S, KEY_HTTP_QUERYSTRING ) );
SC_STR_FUNCTION( sc_r, Q( S, KEY_HTTP_CLF ) );
SC_STR_FUNCTION( sc_R, HTTPUtil_rcNumberToName(
		tContext.tPIHTTP.iLastResultCode ) );
SC_NUM_FUNCTION( sc_s, tContext.tPIHTTP.iStatus, "%d" );
SC_STR_FUNCTION( sc_S, HTTPCore_getServerStamp() );
SC_STR_FUNCTION( sc_u, R( S, KEY_INT_REMOTEUSER ) );
SC_STR_FUNCTION( sc_U, Q( S, KEY_HTTP_URI ) );
SC_STR_FUNCTION( sc_v, HTTPUtil_getHostName(&(tContext.tPIHTTP)) );
SC_STR_FUNCTION( sc_x, R( S, KEY_INT_AUTHTYPE ) );
SC_NUM_FUNCTION( sc_X, tContext.tPIHTTP.iDepth, "%d" );
SC_STR_FUNCTION( sc_y, R( F, KEY_HTTP_CONTENTLENGTH ) );
SC_STR_FUNCTION( sc_Y, Q( F, KEY_HTTP_CONTENTLENGTH ) );
SC_STR_FUNCTION( sc_z, R( S, KEY_INT_SCRIPTNAME ) );
SC_STR_FUNCTION( sc_Z, R( S, KEY_INT_PATHTRANSLATED ) );

/* ---
	Functions which don't conform to the standard template
--- */
static int sc_E( Context &tContext, char *pszBuffer, int iLength ) 
{ 
	/* ---
	Print out all error messages	
	--- */
	int iTotalLen = 0;
	PIDBIterator *pIter = PIDB_getIterator( tContext.tPIHTTP.pResponseDB,
		PIDBTYPE_STRING, KEY_INT_ERRORMESSAGE, 0 );
	if ( pIter )
		{
		for( ;
			PIDBIterator_atValidElement( pIter );
			PIDBIterator_next( pIter ) )
			{
			const char *pDummy;
			const char *pMessage = (const char *)PIDBIterator_current( 
				pIter, &pDummy );
			int iLen = strlen( pMessage );	
			iTotalLen += iLen;
			if ( iLen>iLength ) 
				{ iLen = iLength; };
			if ( iLen )
				{
				assert( pszBuffer );
				strncpy( pszBuffer, pMessage, iLen );
				iLength -= iLen;
				pszBuffer = &(pszBuffer[iLen]);
				};
			};
		PIDBIterator_delete( pIter );
		};

	return iTotalLen;
}

static int sc_g( Context &tContext, char *pszBuffer, int iLength ) 
{ 
	int iOn = (int)PIDB_lookup( tContext.tPIHTTP.pConnectionDB, PIDBTYPE_STRING,
		KEY_HTTPS_KEYSIZE, 0 );
	const char *pString;
	int iValueLen;
	if ( iOn )
		{
		pString = "on";
		iValueLen = 2;
		}
	else
		{
		pString = "off";
		iValueLen = 3;
		};
	int iLen = iValueLen < iLength ? iValueLen : iLength;
	if ( iLen>0 )
		{
		assert( pszBuffer ); 
		strncpy( pszBuffer, pString, iLen ); 
		};
	return iValueLen; 
}

static int sc_t( Context &tContext, char *pszBuffer, int iLength ) 
{ 
	(void)tContext; 
	enum { BUF_SIZE=255 }; 
	char szBuf[BUF_SIZE+1]; 
	
	PIPlatform_beforeUnsafeBlock();
	time_t tT;
	time( &tT );
#if 0
	struct tm *pTms = gmtime( &tT );
	int iValueLen = strftime( szBuf, BUF_SIZE, "%d/%b/%Y:%H:%M:%S +0000", pTms );
#endif
	struct tm *pTms = gmtime( &tT );
	/*
	** Bit of wierd logic here, but were taking into account timezone
	** adjustment on new years eve. Hence use of year. The year multiplier
	** doesn't matter once its greater than 365 (its only used for
	** comparision between gmt and localtime, and they are both multiplyied
	** by the same number
	*/
	int iGmtYearDay = ( pTms->tm_year * 366 ) + pTms->tm_yday;
	int iGmtHourMin = ( pTms->tm_hour * 60 ) + pTms->tm_min ;
	pTms = localtime( &tT );
	int iValueLen = strftime( szBuf, BUF_SIZE, "%d/%b/%Y:%H:%M:%S", pTms );
	int iYearDay = ( pTms->tm_year * 366 ) + pTms->tm_yday;
	int iHourMin = ( pTms->tm_hour * 60 ) + pTms->tm_min ;
	PIPlatform_afterUnsafeBlock();

	int iAdjust = ( iHourMin - iGmtHourMin ) ;
	if ( iGmtYearDay>iYearDay )
		{ iAdjust -= ( 24 * 60 ) ; };
	char cSign;
	if ( iAdjust > 0 )
		{
		cSign = '+';
		}
	else
		{
		cSign = '-';
		iAdjust = -iAdjust;
		};
	sprintf( &(szBuf[20]), " %c%02d%02d", cSign, iAdjust / 60, iAdjust % 60 );
	iValueLen += 6;
	
	int iLen = iValueLen < iLength ? iValueLen : iLength;	
	if ( iLen>0 ) 
		{ 
		assert( pszBuffer ); 
		strncpy( pszBuffer, szBuf, iLen ); 
		}; 
	return iValueLen; 
}

static int sc_T( Context &tContext, char *pszBuffer, int iLength ) 
{ 
	(void)tContext; 
	enum { BUF_SIZE=63 }; 
	char szBuf[BUF_SIZE+1]; 

	long lSeconds;	
	long lMicroSeconds;	
	PITimer_elapsed( tContext.tPIHTTP.pTimer, &lSeconds, &lMicroSeconds );
	sprintf( szBuf, "%ld.%06ld", (long)lSeconds, (long)lMicroSeconds );
	int iValueLen = strlen( szBuf );
	int iLen = iValueLen < iLength ? iValueLen : iLength;
	if ( iLen>0 )
		{
		assert( pszBuffer ); 
		strncpy( pszBuffer, szBuf, iLen ); 
		};
	return iValueLen; 
}

#undef S
#undef F
#undef C
#undef Q
#undef R

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void Internal_InitShortcuts()
{
	aFunctions['a'] = sc_a;	/* administrator */
	aFunctions['A'] = sc_A;	/* IP address of remote client */
	aFunctions['b'] = sc_b;	/* bytes sent */
	aFunctions['c'] = sc_c;	/* content-type */
	aFunctions['C'] = sc_C;	/* content-type - incoming */
	aFunctions['d'] = sc_d;	/* debugging flag */
	aFunctions['D'] = sc_D;	/* handler time delta */
	aFunctions['E'] = sc_E;	/* error messages */
	aFunctions['f'] = sc_f;	/* filename */
	aFunctions['g'] = sc_g;	/* Using SSL?, g='Gaurded'?, */
	aFunctions['G'] = sc_G;	/* SSL secret key size (g='Geheimnis') */
	aFunctions['h'] = sc_h;	/* remote host */
	aFunctions['H'] = sc_H;	/* request HTTP protocol */
	aFunctions['i'] = sc_i;	/* remote ident */
	aFunctions['I'] = sc_I;	/* path info section */
	aFunctions['k'] = sc_k;	/* thread id */
	aFunctions['K'] = sc_K;	/* SSL keysize */
	aFunctions['l'] = sc_l;	/* local address */
	aFunctions['m'] = sc_m;	/* request method from client */
	aFunctions['M'] = sc_M;	/* end of line sequence */
	aFunctions['n'] = sc_n;	/* name of current handler object */
	aFunctions['N'] = sc_N;	/* name of current handler phase */
	aFunctions['o'] = sc_o;	/* object name of virtual host */
	aFunctions['p'] = sc_p;	/* server port */
	aFunctions['P'] = sc_P;	/* process id */	
	aFunctions['q'] = sc_q;	/* query string from client */
	aFunctions['r'] = sc_r;	/* request line */
	aFunctions['R'] = sc_R;	/* name of result code from last handler */
	aFunctions['s'] = sc_s;	/* status */
	aFunctions['S'] = sc_S;	/* server stamp */
	aFunctions['t'] = sc_t;	/* time in CLF */
	aFunctions['T'] = sc_T;	/* time taken to service request */
	aFunctions['u'] = sc_u;	/* remote user */
	aFunctions['U'] = sc_U;	/* URL requested */
	aFunctions['v'] = sc_v;	/* server name */
	aFunctions['x'] = sc_x;	/* authentication type */
	aFunctions['X'] = sc_X;	/* depth of request in subrequest hierarchy */
	aFunctions['y'] = sc_y;	/* content-length : outgoing */
	aFunctions['Y'] = sc_Y;	/* content-length : incoming */
	aFunctions['z'] = sc_z;	/* script name */
	aFunctions['Z'] = sc_Z;	/* path translated */

	iShortcutsInitialized = 1;
}
	
/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class ShortcutComponent : public ComponentBase
{
private:
	SCFnWrite fnWrite;

public:
	ShortcutComponent( char c )
	:	fnWrite( 0 )
		{
		if ( !iShortcutsInitialized )
			{
	  		Internal_InitShortcuts();
			};
		fnWrite = aFunctions[c];
		};
	
	virtual void WriteDebugTree( int iIndent ) const
		{
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "Shortcut:" << endl;
		};

	virtual int Write( Context &tContext, char *pszBuffer, int iLength ) const
		{
		if ( !fnWrite ) 
			{ return 0; };
		return (fnWrite)( tContext, pszBuffer, iLength );
		};

	virtual int IsOK() const { return fnWrite!=0; };
	virtual int IsStatic() const	{ return 0; /* some sc's only */ };
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Token class. Encapsulates a token in the input stream.
\*____________________________________________________________________________*/
class Token
{
public:
	enum Type {
		DOLLAR=1,
		PERCENT,
		OPEN_PAREN,
		CLOSE_PAREN,
		COMMA,
		AMPERSAND,
		NUMBER,
		IDENTIFIER,
		LITERAL,
		OTHER
		};
	Type eType;
	PIString sStr;
	char cChar;
	long lVal;

	/* --- constructors --- */
	Token( const Token &tT )
		: eType( tT.eType ), sStr( tT.sStr ),
		cChar( tT.cChar ), lVal( tT.lVal )
		{};
	Token( Type eTheType, char c ) : eType( eTheType ), cChar( c ), lVal( 0 )
		{};
	Token( char c ) : eType( OTHER ), cChar( c ), lVal( 0 )
		{};
	Token( Type eTheType, const PIString &sTheString )
		: eType( eTheType ), sStr( sTheString ), cChar( '\0' ), lVal( 0 )
		{ 
		const char *pS = sStr;
		cChar = pS ? *pS : '\0';
		};

	/* --- descriptions --- */
	void GetDescription( PIString &sDescription )
		{
		switch( eType )
			{
			case DOLLAR: 
			case PERCENT: 
			case OPEN_PAREN: 
			case CLOSE_PAREN: 
			case COMMA: 
			case AMPERSAND: 
			case OTHER: 
				{
				char szX[] = "X";
				*szX = cChar;
				sDescription = szX;
				break;
				};

			case NUMBER:		sDescription = "Number"; break;
			case IDENTIFIER:	sDescription = "Identifier"; break;
			default:
				assert( 0 ); sDescription = "Unknown";
			};
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
class TokenSource
{
private:
	PIString sExpression;
	int iPos;
	Token *pPushedBack;		
	
public:
	TokenSource( const char *pTheExpression )
	:	sExpression( pTheExpression ),
		iPos( 0 ),
		pPushedBack( 0 )
		{
		};

	~TokenSource()
		{
		PI_DELETE( pPushedBack );
		};

	inline const PIString &GetRawString()	{ return sExpression; };
	inline int GetPosition()				{ return iPos; };

private:
	void Internal_PushBack()
		{ assert( iPos>0 ); iPos--; };

	Token *Internal_GetNextToken( PIString &sError )
		{
		enum {
			NORMAL=1,
			ESCAPED,
			QUOTED,
			COMMENT_START,
			IN_COMMENT,
			COMMENT_END,
			IDENTIFIER
			} eState = NORMAL;

		/* --- persistent variables --- */
		PIString sBuffer;

		/* --- lexical state machine --- */
		for(;;)
			{
			if ( sExpression.Len()==iPos )
				{
				switch( eState )
					{
					case NORMAL:	return 0;
					case COMMENT_START:	return PI_NEW( Token( '/' ) );
					case ESCAPED:	return PI_NEW( Token( '\\' ) );
					case IDENTIFIER:
						return PI_NEW(Token( Token::IDENTIFIER, sBuffer ));
					case QUOTED:
						sError = "End of expression reading quoted literal.";
						return 0;
					case IN_COMMENT:
					case COMMENT_END:
						sError = "End of expression in reading comment.";
						return 0;
					default: assert( 0 ); return 0;
					};
				};
			const char *pTmp = sExpression;
			char c = pTmp[iPos++];

			switch( eState )
				{
				case NORMAL:
					if ( c=='\\' )
						{ eState = ESCAPED; break; }
					else if ( (isalpha(c)) || (c=='_') )
						{
						char szX[] = "X";
						*szX = c;
						sBuffer = szX;
						eState = IDENTIFIER;
						break;
						};
					switch( c )
						{
						case '\\': eState = ESCAPED; break;
						case '\'': eState = QUOTED; break;
						case '/': eState = COMMENT_START; break;
						case '%': return PI_NEW( Token(Token::PERCENT, '%'));
						case '$': return PI_NEW( Token(Token::DOLLAR, '$') );
						case '(': return PI_NEW( Token(Token::OPEN_PAREN,'(') );
						case ')': return PI_NEW( Token(Token::CLOSE_PAREN,')'));
						case ',': return PI_NEW( Token(Token::COMMA,',') );
						case '&': return PI_NEW( Token(Token::AMPERSAND,'&') );
						default:;
							return PI_NEW( Token( c ) );
						};
					break;

				case COMMENT_START:
					if ( c=='*' )
						{ eState = IN_COMMENT; break; };
					Internal_PushBack();	
					return PI_NEW( Token( '/' ) );

				case IN_COMMENT:
					if ( c=='*' )
						{ eState = COMMENT_END; break; };
					break;

				case COMMENT_END:
					if ( c=='/' )
						{ eState = NORMAL; break; };
					eState = IN_COMMENT;
					break;

				case IDENTIFIER:
					if ( !(isalnum(c) || (c=='_')) )
						{
						eState = NORMAL;
						Internal_PushBack();
						return PI_NEW(Token( Token::IDENTIFIER, sBuffer ));
						};
					sBuffer.Concatenate( c );
					break;

				case ESCAPED:
					eState = NORMAL;
					switch( c )
						{
						case '\\': c = '\\'; break;
						case 'a': c = '\a'; break;
						case 'b': c = '\b'; break;
						case 'f': c = '\f'; break;
						case 'n': c = '\n'; break;
						case 'r': c = '\r'; break;
						case 't': c = '\t'; break;
						case 'v': c = '\v'; break;
						default:;
						};
					return PI_NEW( Token( c ) );

				case QUOTED:
					if ( c=='\'' )
						{
						eState = NORMAL;
						return PI_NEW(Token( Token::LITERAL, sBuffer ));
						};
					sBuffer.Concatenate( c );
					break;

				default: assert( 0 ); return 0;
				};
			};

		/* --- should never get here --- */
		assert( 0 );
		return 0;
		};

public:
	inline Token *GetNextToken( PIString &sError )
		{ 
		if ( pPushedBack )
			{
			Token *pTmp = pPushedBack;
			pPushedBack = 0;
			return pTmp;
			};
		return Internal_GetNextToken( sError );
		};

	inline void PushBack( Token *pToPushBack )
		{
		assert( !pPushedBack );
		pPushedBack = pToPushBack;
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int Internal_ExpressionToString( Context &tContext,
	_Pi3Expression *pExpr, char *pszBuffer, int iLength )
{
	return pExpr->Write( tContext, pszBuffer, iLength );
}

/*____________________________________________________________________________*\
 *
 Class:
 Description:
		- make this more efficient by having functions be objects.
\*____________________________________________________________________________*/
class FunctionComponent : public ComponentBase
{
protected:
	/* ---
	encapsulation of a function arguments
	--- */
	class FunctionArgument
	{
	public:
		enum { STRING=0, NUMBER, EXPRESSION } eType;
		union {	
			PIString *pString;
			int iNumber;
			_Pi3Expression *pExpr;
			};

		FunctionArgument( PIString *pTheString )
		:	eType( STRING ), pString( pTheString )	{ assert( pString ); };

		FunctionArgument( int iTheNumber )
		:	eType( NUMBER ), iNumber( iTheNumber )	{};

		FunctionArgument( _Pi3Expression *pTheExpr )
		:	eType( EXPRESSION ), pExpr( pTheExpr )	{ assert( pExpr ); };

		~FunctionArgument()
			{
			switch( eType )
				{
				case STRING:
					PI_DELETE( pString ); break;
				case NUMBER:
					break;
				case EXPRESSION:
					PI_DELETE( pExpr ); break;
				default:
					assert( 0 );
				};
			};

		int Write( Context &tContext, char *pszBuffer, int iLength ) const
			{
			switch( eType )
				{
				case STRING: {
					int iValueLen = pString->Len();
					int iLen = iValueLen>iLength ? iLength : iValueLen;
					if ( iLen>0 && pszBuffer )
						{ strncpy( pszBuffer, *pString, iLen ); };
					return iValueLen;
					};

				case NUMBER: {
					enum { NUM_BUF=63 };
					char szBuf[NUM_BUF+1];
					sprintf( szBuf, "%d", iNumber );
					int iValueLen = strlen( szBuf );
					int iLen = iValueLen>iLength ? iLength : iValueLen;
					strncpy( pszBuffer, szBuf, iLen );
					return iValueLen;
					};
					
				case EXPRESSION:
					return Internal_ExpressionToString(
						tContext, pExpr, pszBuffer, iLength );

				default:
					assert( 0 );
				};
			assert( 0 );
			return -1;
			};

		void WriteDebugTree( int iIndent ) const
			{
			switch( eType )
				{
				case STRING:
					cerr << "STRING:<" << *pString << ">";
					break;

				case NUMBER:
					cerr << "NUMBER:<" << iNumber << ">" << endl;
					break;

				case EXPRESSION:
					cerr << endl;
					pExpr->WriteDebugTree( iIndent+1 );
					break;

				default:
					assert( 0 );
				};
			};
	};

	int iOK;	

	/* --- some useful methods --- */
	void ExpressionToNumber( Context &tContext,
			_Pi3Expression *pExpr, int &iResult ) const
		{
		enum { BUF_SIZE=63 };
		char szBuf[BUF_SIZE+1];
		szBuf[BUF_SIZE]='\0';

		register int iNumberLength =
			pExpr->Write( tContext, szBuf, BUF_SIZE );

		iResult = iNumberLength>0 ? atoi( szBuf ) : iNumberLength;
		};

	void FunctionArgumentToNumber( Context &tContext,
			FunctionArgument *pArg, int &iResult ) const
		{
		if ( pArg->eType==FunctionArgument::NUMBER )
			{
			iResult = pArg->iNumber;
			}
		else
			{
			enum { BUF_SIZE=63 };
			char szBuf[BUF_SIZE+1];
			*szBuf='\0';
			int iNumberLength = pArg->Write( tContext, szBuf, BUF_SIZE );
			assert( iNumberLength!=-1 );

			iResult = iNumberLength>0 ? atoi( szBuf ) : iNumberLength;
			};
		};

	int FunctionArgumentToString( Context &tContext,
		FunctionArgument *pArg, char *pszBuffer, int iLength ) const
		{
		return pArg->Write( tContext, pszBuffer, iLength );
		};

	FunctionArgument *ExpressionToFunctionArgument( _Pi3Expression *pExpr )
		{
		if ( !pExpr )
			{ return 0; };

		if ( !pExpr->IsStatic() )
			{ return PI_NEW( FunctionArgument( pExpr ) ); };

		/* --- dummy --- */
		PIHTTP tPIHTTP( (PIDB *)0 );
		Context tContext( 0, tPIHTTP, 0 );
		int iLen = pExpr->Write( tContext, 0, 0 );
		if ( iLen==-1 )
			{
			assert( 0 );
			PI_DELETE( pExpr );
			return 0; 
			};
		if ( !iLen )
			{
			PI_DELETE( pExpr );
			return PI_NEW( FunctionArgument( PI_NEW( PIString ) ) );
			}

		char *pBuffer = PI_NEW( char[iLen+1] );
		assert( pBuffer );
		iLen = pExpr->Write( tContext, pBuffer, iLen );

		/* --- done with expression --- */
		PI_DELETE( pExpr );

		FunctionArgument *pArg = PI_NEW( FunctionArgument( 
			PI_NEW( PIString( pBuffer, iLen ) ) ) );

		PI_DELETE( [] pBuffer );

		return pArg;
		};

public:
	FunctionComponent()
	:	iOK( 0 )
		{};

	virtual ~FunctionComponent()
		{};
	
	virtual int IsOK() const { return iOK; };

	/* --- pure virtual functions for derived classes --- */
	virtual const char *GetName() const=0;
	virtual int Min()=0;
	virtual int Max()=0;

	/* ---
	Optimize expression. Return 1 on success, 0 on error
	--- */
	virtual int Compile( DblList &lExpressions, PIString &sError )=0;

	/* --- public functions --- */
	static FunctionComponent *MakeFunctionComponent( PIString &sFunctionName,
		TokenSource &tSource, FnPi3Write *pFunctions,
		PIString &sError, int &iNestedMessage );
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class TruncFunctionComponent : public FunctionComponent
{
private:
	FunctionArgument *pString;
	FunctionArgument *pLength;

public:
	TruncFunctionComponent()
	:	FunctionComponent(),
		pString( 0 ),
		pLength( 0 )
		{};

	virtual ~TruncFunctionComponent()
		{
 		PI_DELETE( pString );		
		PI_DELETE( pLength );
		};

	virtual const char *GetName() const		{ return "trunc"; };

	/* --- */
	virtual void WriteDebugTree( int iIndent ) const
		{
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "Function:'" << GetName() << "'";
		if ( IsStatic() )
			{ cerr << "<Static>"; }
		else
			{ cerr << "<Dynamic>"; }
		if ( !pString )
			{ cerr << endl; return; };
		pString->WriteDebugTree( iIndent );
		assert( pLength );
		cerr << ", ";
		pLength->WriteDebugTree( iIndent );
		};

	/* --- */
	virtual int Write( Context &tContext, char *pszBuffer, int iLength ) const
		{
		if ( !pString )
			{ return 0; };	/* --- string is empty --- */

		assert( pLength );

		int iTruncLength;
		if ( pLength->eType==FunctionArgument::NUMBER )
			{ iTruncLength = pLength->iNumber; }
		else
			{
			assert( pLength->eType==FunctionArgument::EXPRESSION );

			ExpressionToNumber( tContext, pLength->pExpr, iTruncLength );
			};

		if ( iTruncLength<=0 )
			{
			if ( !iTruncLength )
				{ return 0; };
			/* else	iTruncLength==-1 */
			assert( 0 );
			const char pError[] = "_Pi3Expression::trunc(): invalid length";
			strncpy( pszBuffer, pError, iLength );
			return ((int)sizeof( pError ));
			};

		/* --- 
		Write String
		--- */
		if ( pString->eType==FunctionArgument::EXPRESSION )
			{
			int iRet = pString->pExpr->Write( tContext, pszBuffer,
				iTruncLength );
			if ( iRet==-1 )
				{
				assert( 0 );
				return -1;
				};
			if ( iRet<iTruncLength )
				{ return iRet; };
			return iTruncLength;
			}
		else
			{
			assert( pString->eType==FunctionArgument::STRING );

			PIString &sString = *( pString->pString );
			assert( sString.Len() );
			if ( sString.Len()<iTruncLength )
				{ iTruncLength = sString.Len(); };
			strncpy( pszBuffer, sString, iTruncLength );
			return iTruncLength;
			};
		};

	virtual int Min()	{ return 2; };
	virtual int Max()	{ return 2; };

	/* ---
	Optimize expression. Return 1 on success, 0 on error
	--- */
	virtual int Compile( DblList &lExpressions, PIString &sError )
		{
		DblListIterator i( lExpressions );
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pEString = (_Pi3Expression *)i.Current();
		i++;
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pELength = (_Pi3Expression *)i.Current();

		/* --- make dummy PIHTTP object --- */
		PIHTTP tPIHTTP( (PIDB *)0 );
		Context tContext( 0, tPIHTTP, 0 );

		/* --- clear the list --- */
		assert( lExpressions.Size()==Min() );
		assert( lExpressions.Size()==Max() );
		lExpressions.Clear();

		/* --- set string argument --- */
		if ( !pEString->IsStatic() )
			{
			pString = PI_NEW( FunctionArgument( pEString ) );
			}
		else
			{	
			int iLength = pEString->Write( tContext, 0, 0 );
			if ( iLength<0 )
				{
				assert( 0 );
				return -1;
				};
			if ( iLength==0 )
				{
				PI_DELETE( pEString );
				pEString = 0;
				}
			else
				{
				char *pBuffer = PI_NEW( char[iLength+1] );
				pEString->Write( tContext, pBuffer, 0 );
				pBuffer[iLength] = '\0';
				pString = PI_NEW( FunctionArgument(
					PI_NEW( PIString( pBuffer, iLength ) ) ) );
				PI_DELETE( [] pBuffer );
				};
			};

		/* --- set length argument, if string is non-empty --- */
		if ( pString )
			{
			if ( !pELength->IsStatic() )
				{
				pLength = PI_NEW( FunctionArgument( pELength ) );
				}
			else
				{
				/* --- preprocess expression to a number --- */
				int iNumber=-1;
				ExpressionToNumber( tContext, pELength, iNumber );
				PI_DELETE( pELength );
				if ( iNumber==-1 )
					{
					sError = "_Pi3Expression::trunc(): invalid length argument";
					return 0;
					};
				pLength = PI_NEW( FunctionArgument( iNumber ) );
				};
			};

		/* --- done --- */
		iOK = 1;
		return 1;
		};

	virtual int IsStatic() const
		{
		if ( !pString )
			{ return 1; };	/* --- static because its empty --- */

		assert( pLength );

		return pString->eType==FunctionArgument::STRING && 
			pLength->eType==FunctionArgument::NUMBER;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	This function should be re-written.
\*____________________________________________________________________________*/
class AlignFunctionComponent : public FunctionComponent
{
private:
	FunctionArgument *pString;
	FunctionArgument *pLength;
	FunctionArgument *pFlags;
	FunctionArgument *pPadding;

public:
	AlignFunctionComponent()
	:	FunctionComponent(),
		pString( 0 ),
		pLength( 0 ),
		pFlags( 0 ),
		pPadding( 0 )
		{};

	virtual ~AlignFunctionComponent()
		{
 		PI_DELETE( pString );		
		PI_DELETE( pLength );
		PI_DELETE( pFlags );
		PI_DELETE( pPadding );
		};

	virtual const char *GetName() const		{ return "align"; };

	/* --- */
	virtual void WriteDebugTree( int iIndent ) const
		{
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "Function:'" << GetName() << "'";
		if ( IsStatic() )
			{ cerr << "<Static>"; }
		else
			{ cerr << "<Dynamic>"; }
		assert( pString );
		pString->WriteDebugTree( iIndent );
		cerr << ", ";
		assert( pFlags );
		pFlags->WriteDebugTree( iIndent );
		cerr << ", ";
		assert( pPadding );
		pPadding->WriteDebugTree( iIndent );
		};

	/* --- */
	virtual int Write( Context &tContext, char *pszBuffer, int iLength ) const 
		{
		assert( pString );
		assert( pLength );
		assert( pFlags );
		assert( pPadding );
		assert( pszBuffer || !iLength );

		/* --- get field length --- */
		int iFieldLength;			/* the length of the field to write */
		FunctionArgumentToNumber( tContext, pLength, iFieldLength );

		/* --- give error if field length is wrong --- */
		if ( iFieldLength<=0 )
			{
			if ( !iFieldLength )
				{
				return 0;
				};
			/* else	iFieldLength==-1 */
			const char pError[] = "_Pi3Expression::align(): invalid length";
			if ( iLength>0 )
				{ strncpy( pszBuffer, pError, iLength ); };
			return ((int)sizeof( pError ));
			};

		enum { BUF_SIZE=127 };

		/* --- get padding --- */
		char szPaddingBuf[BUF_SIZE+1];
		int iPaddingLen = pPadding->Write( tContext, szPaddingBuf,
			BUF_SIZE );
		if ( iPaddingLen==-1 )
			{ return -1; };
		if ( iPaddingLen>BUF_SIZE )
			{ iPaddingLen=BUF_SIZE; };
		PIString sPadding;
		if ( iPaddingLen>0 )
			{
			szPaddingBuf[iPaddingLen]='\0';
			sPadding = szPaddingBuf;
			}
		else
			{
			sPadding = "                    ";
			iPaddingLen = sPadding.Len();
			};

		/* --- get flags --- */
		char szFlagsBuf[BUF_SIZE+1];
		int iFlagsLen = pFlags->Write( tContext, szFlagsBuf, BUF_SIZE );
		if ( iFlagsLen==-1 )
			{ return -1; };
		if ( iFlagsLen>BUF_SIZE )
			{ iFlagsLen = BUF_SIZE; };
		char *pszFlagsBuf = szFlagsBuf;
		/* --- eat leading whitespace --- */
		while( iFlagsLen>0 && (isspace(*pszFlagsBuf)) )
			{
			iFlagsLen--;
			pszFlagsBuf++;
			};
		if ( iFlagsLen>0 )
			{ pszFlagsBuf[iFlagsLen] = '\0'; };

		/* --- write in string --- */
		int iStringLen = pString->Write( tContext, pszBuffer, iLength );
		if ( iStringLen==-1 )
			{ return -1; };
		if (
			iStringLen>=iFieldLength ||
			iStringLen>=iLength )
			{ return iStringLen; };
		assert( iPaddingLen );

#if 0
/*
** Centering disabled, because this code is not right - REDO
**
** NOTE: 'center' has also been removed from the documentation
*/
		/* --- align the field correctly --- */
		if ( iFlagsLen>0 && !PIUtil_strncmpi( pszFlagsBuf, "center", 6 ) )
			{
			int iInitialPadding = ( iFieldLength - iStringLen ) / 2;
			int iEndPadding = ( iFieldLength - iStringLen ) - iInitialPadding;
			if ( iInitialPadding )
				{
				if ( iInitialPadding<iLength )
					{
					int iTheLen = iStringLen;
					if ( iTheLen>(iLength-iInitialPadding) )
						{ iTheLen = (iLength - iInitialPadding ); };
					memmove( &(pszBuffer[iInitialPadding]), pszBuffer,
						iTheLen );
					};
				};

			const char *pPadding = sPadding;

			int iPos;
			for( iPos=0; iPos<iInitialPadding; )
				{ 
				int iTheLen = iInitialPadding - iPos;
				if ( iTheLen>iPaddingLen )
					{ iTheLen = iPaddingLen; };
				if ( iPos<iLength )	
					{
					if ( iTheLen<(iLength-iPos) )
						{ iTheLen = (iLength - iPos ); };
					strncpy( &(pszBuffer[iPos]), pPadding, iTheLen );
					};
				iPos+=iTheLen;
				}
			pszBuffer = &( pszBuffer[iInitialPadding+iStringLen] );
			for( iPos=0; iPos<iEndPadding; )
				{ 
				int iTheLen = iEndPadding - iPos;
				if ( iTheLen>iPaddingLen )
					{ iTheLen = iPaddingLen; };
				strncpy( &(pszBuffer[iPos]), pPadding, iTheLen );
				iPos+=iTheLen;
				}

			return iFieldLength;
			}
#endif
		if ( iFlagsLen>0 && !PIUtil_strncmpi( pszFlagsBuf, "right", 5 ) )
			{
			int iTmp = iFieldLength - iStringLen;
	
			/* --
			Move text to the right
			--- */
			if ( iTmp<iLength )
				{
				int iTheLen = iStringLen;
				if ( (iLength-iTmp)<iTheLen )
					{ iTheLen = (iLength-iTmp); };
				memmove( &(pszBuffer[iTmp]), pszBuffer, iTheLen );
				};

			/* ---
			Write in the padding
			--- */
			const char *pPadding = sPadding;
			if ( iTmp>iLength )
				{ iTmp = iLength; };
			for( int iPos=0; iPos<iTmp; )
				{ 
				int iTheLen = iTmp - iPos;
				if ( iTheLen>iPaddingLen )
					{ iTheLen = iPaddingLen; };
				strncpy( &(pszBuffer[iPos]), pPadding, iTheLen );
				iPos += iTheLen;
				}
			return iFieldLength;
			}
		else /* default is left */
			{
			/* --- real field size to fill --- */
			int iToFill = iFieldLength < iLength ? iFieldLength : iLength;

			/* --- real number of spaces to pad --- */
			int iToPad = iToFill - iStringLen;

			assert( iPaddingLen );
			if ( iToPad>0 && (iPaddingLen>0)/* sanity */ )
				{
				pszBuffer = &( pszBuffer[iStringLen] );
				const char *pPadding = sPadding;
				for( int iPos=0; iPos<iToPad; )
					{
					int iTheLen = iToPad - iPos;
					if ( iTheLen>iPaddingLen )
						{ iTheLen = iPaddingLen; };
					strncpy( &(pszBuffer[iPos]), pPadding, iTheLen );
					iPos += iTheLen;
					};
				};
			return iFieldLength;
			};
		};

	virtual int Min()	{ return 2; };
	virtual int Max()	{ return 4; };

	/* ---
	Optimize expression. Return 1 on success, 0 on error
	--- */
	virtual int Compile( DblList &lExpressions, PIString & /* sError */ )
		{
		DblListIterator i( lExpressions );
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pEString = (_Pi3Expression *)i.Current();
		i++;
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pELength = (_Pi3Expression *)i.Current();
		_Pi3Expression *pEFlags;
		_Pi3Expression *pEPadding;
		i++;
		if ( i.BadIndex() )
			{ pEFlags = PI_NEW( _Pi3Expression() ); }
		else
			{
			pEFlags = (_Pi3Expression *)i.Current();
			i++;
			};
		if ( i.BadIndex() )
			{ pEPadding = PI_NEW( _Pi3Expression() ); }
		else
			{ pEPadding = (_Pi3Expression *)i.Current(); };

		/* --- clear the list --- */
		assert( lExpressions.Size()>=Min() );
		assert( lExpressions.Size()<=Max() );
		lExpressions.Clear();

		pString = ExpressionToFunctionArgument( pEString );
		pLength = ExpressionToFunctionArgument( pELength );
		pFlags = ExpressionToFunctionArgument( pEFlags );
		pPadding = ExpressionToFunctionArgument( pEPadding );

		if ( !pString || !pLength || !pFlags || !pPadding )
			{ return 0; };

		if ( pLength->eType==FunctionArgument::STRING )
			{
			FunctionArgument *pSaveLength = pLength;
			pLength = PI_NEW( FunctionArgument( atoi(*(pLength->pString)) ) );
			PI_DELETE( pSaveLength );
			};

		/* --- done --- */
		iOK = 1;
		return 1;
		};

	virtual int IsStatic() const
		{
		assert( pString );
		assert( pLength );
		assert( pFlags );
		assert( pPadding );

		return pString->eType==FunctionArgument::STRING && 
			pLength->eType==FunctionArgument::NUMBER &&
			pFlags->eType==FunctionArgument::STRING && 
			pPadding->eType==FunctionArgument::STRING;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class IfFunctionComponent : public FunctionComponent
{
private:
	FunctionArgument *pBoolean;
	FunctionArgument *pOnTrue;
	FunctionArgument *pOnFalse;

public:
	IfFunctionComponent()
	:	FunctionComponent(),
		pBoolean( 0 ),
		pOnTrue( 0 ),
		pOnFalse( 0 )
		{};

	virtual ~IfFunctionComponent()
		{
 		PI_DELETE( pBoolean );		
		PI_DELETE( pOnTrue );
		PI_DELETE( pOnFalse );
		};

	virtual const char *GetName() const		{ return "if"; };

	/* --- */
	virtual void WriteDebugTree( int iIndent ) const
		{
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "Function:'" << GetName() << "'";
		if ( IsStatic() )
			{ cerr << "<Static>"; }
		else
			{ cerr << "<Dynamic>"; }
		assert( pBoolean );
		pBoolean->WriteDebugTree( iIndent );
		cerr << ", ";
		assert( pOnTrue );
		pOnTrue->WriteDebugTree( iIndent );
		cerr << ", ";
		assert( pOnFalse );
		pOnFalse->WriteDebugTree( iIndent );
		};

	/* --- */
	virtual int Write( Context &tContext, char *pszBuffer, int iLength ) const
		{
		assert( pBoolean );
		assert( pOnTrue );
		assert( pOnFalse );

		/* --- switch on the length of pBoolean, to write out --- */
		switch( pBoolean->Write( tContext, 0, 0 ) )
			{
			case -1:	return -1;
			case 0:		return pOnFalse->Write( tContext, pszBuffer, iLength );
			default:	return pOnTrue->Write( tContext, pszBuffer, iLength );
			};
		};

	virtual int Min()	{ return 2; };
	virtual int Max()	{ return 3; };

	/* ---
	Optimize expression. Return 1 on success, 0 on error
	--- */
	virtual int Compile( DblList &lExpressions, PIString & /* sError */ )
		{
		DblListIterator i( lExpressions );
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pEBoolean = (_Pi3Expression *)i.Current();
		i++;
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pEOnTrue = (_Pi3Expression *)i.Current();
		_Pi3Expression *pEOnFalse = 0;
		i++;
		if ( i.BadIndex() )
			{ pEOnFalse = PI_NEW( _Pi3Expression() ); }
		else
			{
			pEOnFalse = (_Pi3Expression *)i.Current();
			i++;
			};

		/* --- clear the list --- */
		assert( lExpressions.Size()>=Min() );
		assert( lExpressions.Size()<=Max() );
		lExpressions.Clear();

		pBoolean = ExpressionToFunctionArgument( pEBoolean );
		pOnTrue = ExpressionToFunctionArgument( pEOnTrue );
		pOnFalse = ExpressionToFunctionArgument( pEOnFalse );

		if ( !pBoolean || !pOnTrue || !pOnFalse )
			{ return 0; };

		/* --- done --- */
		iOK = 1;
		return 1;
		};

	virtual int IsStatic() const
		{
		assert( pBoolean );
		assert( pOnTrue );
		assert( pOnFalse );

		return pBoolean->eType==FunctionArgument::STRING && 
			pOnTrue->eType==FunctionArgument::STRING &&
			pOnFalse->eType==FunctionArgument::STRING;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class AndFunctionComponent : public FunctionComponent
{
private:
	FunctionArgument *pArg1;
	FunctionArgument *pArg2;

public:
	AndFunctionComponent()
	:	FunctionComponent(),
		pArg1( 0 ),
		pArg2( 0 )
		{};

	virtual ~AndFunctionComponent()
		{
 		PI_DELETE( pArg1 );		
		PI_DELETE( pArg2 );
		};

	virtual const char *GetName() const		{ return "and"; };

	/* --- */
	virtual void WriteDebugTree( int /* iIndent */ ) const
		{
#if 0
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "Function:'" << GetName() << "'";
		if ( IsStatic() )
			{ cerr << "<Static>"; }
		else
			{ cerr << "<Dynamic>"; }
		assert( pBoolean );
		pBoolean->WriteDebugTree( iIndent );
		cerr << ", ";
		assert( pOnTrue );
		pOnTrue->WriteDebugTree( iIndent );
		cerr << ", ";
		assert( pOnFalse );
		pOnFalse->WriteDebugTree( iIndent );
#endif
		};

	/* --- */
	virtual int Write( Context &tContext, char *pszBuffer, int iLength ) const
		{
		assert( pArg1 );
		assert( pArg2 );

		/* --- switch on pArg1 --- */
		switch( pArg1->Write( tContext, 0, 0 ) )
			{
			case -1:	return -1;
			case 0:		return 0;
			default:;
			};

		/* --- switch on pArg2 --- */
		switch( pArg2->Write( tContext, 0, 0 ) )
			{
			case -1:	return -1;
			case 0:		return 0;
			default:;
			};

		if ( pszBuffer && iLength )
			{
			int iLen = ( iLength < 4 ) ? iLength : 4;
			strncpy( pszBuffer, "true", iLen );
			};

		return 4;
		};

	virtual int Min()	{ return 2; };
	virtual int Max()	{ return 2; };

	/* ---
	Optimize expression. Return 1 on success, 0 on error
	--- */
	virtual int Compile( DblList &lExpressions, PIString & /* sError */ )
		{
		DblListIterator i( lExpressions );
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pEArg1 = (_Pi3Expression *)i.Current();
		i++;
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pEArg2 = (_Pi3Expression *)i.Current();
		i++;

		/* --- clear the list --- */
		assert( lExpressions.Size()>=Min() );
		assert( lExpressions.Size()<=Max() );
		lExpressions.Clear();

		pArg1 = ExpressionToFunctionArgument( pEArg1 );
		pArg2 = ExpressionToFunctionArgument( pEArg2 );

		if ( !pArg1 || !pArg2 )
			{ return 0; };

		/* --- done --- */
		iOK = 1;
		return 1;
		};

	virtual int IsStatic() const
		{
		assert( pArg1 );
		assert( pArg2 );

		return pArg1->eType==FunctionArgument::STRING && 
			pArg2->eType==FunctionArgument::STRING;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class OrFunctionComponent : public FunctionComponent
{
private:
	FunctionArgument *pArg1;
	FunctionArgument *pArg2;

public:
	OrFunctionComponent()
	:	FunctionComponent(),
		pArg1( 0 ),
		pArg2( 0 )
		{};

	virtual ~OrFunctionComponent()
		{
 		PI_DELETE( pArg1 );		
		PI_DELETE( pArg2 );
		};

	virtual const char *GetName() const		{ return "and"; };

	/* --- */
	virtual void WriteDebugTree( int /* iIndent */ ) const
		{
#if 0
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "Function:'" << GetName() << "'";
		if ( IsStatic() )
			{ cerr << "<Static>"; }
		else
			{ cerr << "<Dynamic>"; }
		assert( pBoolean );
		pBoolean->WriteDebugTree( iIndent );
		cerr << ", ";
		assert( pOnTrue );
		pOnTrue->WriteDebugTree( iIndent );
		cerr << ", ";
		assert( pOnFalse );
		pOnFalse->WriteDebugTree( iIndent );
#endif
		};

	/* --- */
	virtual int Write( Context &tContext, char *pszBuffer, int iLength ) const
		{
		assert( pArg1 );
		assert( pArg2 );

		/* --- switch on pArg1 --- */
		switch( pArg1->Write( tContext, 0, 0 ) )
			{
			case -1:	return -1;
			case 0:		break;
			default:;
				goto or_succeeded;
			};

		/* --- switch on pArg2 --- */
		switch( pArg2->Write( tContext, 0, 0 ) )
			{
			case -1:	return -1;
			case 0:		break;
			default:;
				goto or_succeeded;
			};

		/* --- OR failed --- */
		return 0;

or_succeeded:
		if ( pszBuffer && iLength )
			{
			int iLen = ( iLength < 4 ) ? iLength : 4;
			strncpy( pszBuffer, "true", iLen );
			};

		return 4;
		};

	virtual int Min()	{ return 2; };
	virtual int Max()	{ return 2; };

	/* ---
	Optimize expression. Return 1 on success, 0 on error
	--- */
	virtual int Compile( DblList &lExpressions, PIString & /* sError */ )
		{
		DblListIterator i( lExpressions );
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pEArg1 = (_Pi3Expression *)i.Current();
		i++;
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pEArg2 = (_Pi3Expression *)i.Current();
		i++;

		/* --- clear the list --- */
		assert( lExpressions.Size()>=Min() );
		assert( lExpressions.Size()<=Max() );
		lExpressions.Clear();

		pArg1 = ExpressionToFunctionArgument( pEArg1 );
		pArg2 = ExpressionToFunctionArgument( pEArg2 );

		if ( !pArg1 || !pArg2 )
			{ return 0; };

		/* --- done --- */
		iOK = 1;
		return 1;
		};

	virtual int IsStatic() const
		{
		assert( pArg1 );
		assert( pArg2 );

		return pArg1->eType==FunctionArgument::STRING && 
			pArg2->eType==FunctionArgument::STRING;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class CmpFunctionComponent : public FunctionComponent
{
private:
	FunctionArgument *pArg1;
	FunctionArgument *pArg2;
	int iCaseSensitive;

public:
	CmpFunctionComponent( int iIsCaseSensitive )
	:	FunctionComponent(),
		pArg1( 0 ),
		pArg2( 0 ),
		iCaseSensitive( iIsCaseSensitive )
		{};

	virtual ~CmpFunctionComponent()
		{
 		PI_DELETE( pArg1 );		
		PI_DELETE( pArg2 );
		};

	virtual const char *GetName() const
		{ return iCaseSensitive?"cmp":"cmpi"; };

	/* --- */
	virtual void WriteDebugTree( int /* iIndent */ ) const
		{
#if 0
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "Function:'" << GetName() << "'";
		if ( IsStatic() )
			{ cerr << "<Static>"; }
		else
			{ cerr << "<Dynamic>"; }
		assert( pBoolean );
		pBoolean->WriteDebugTree( iIndent );
		cerr << ", ";
		assert( pOnTrue );
		pOnTrue->WriteDebugTree( iIndent );
		cerr << ", ";
		assert( pOnFalse );
		pOnFalse->WriteDebugTree( iIndent );
#endif
		};

	/* --- */
	virtual int Write( Context &tContext, char *pszBuffer, int iLength ) const
		{
		assert( pArg1 );
		assert( pArg2 );

		enum { BUF_SIZE=127 };
		char szBuf1[BUF_SIZE+1];
		char szBuf2[BUF_SIZE+2];

		int iLen1 = pArg1->Write( tContext, szBuf1, BUF_SIZE );
		int iLen2 = pArg2->Write( tContext, szBuf2, BUF_SIZE );

		if ( iLen1==-1 || iLen2==-1 )
			{	
			return -1;
			};

		if ( iLen1!=iLen2 )
			{
			return 0;
			};

		char *p1 = szBuf1;
		char *p2 = szBuf2;
		if ( iLen1>BUF_SIZE )
			{
			p1 = PI_NEW( char[iLen1+1] );
			pArg1->Write( tContext, p1, iLen1 );
			};
		p1[iLen1] = '\0';
		if ( iLen2>BUF_SIZE )
			{
			p2 = PI_NEW( char[iLen2+1] );
			pArg2->Write( tContext, p2, iLen2 );
			};
		p1[iLen2] = '\0';

		int iResult = iCaseSensitive ?
				!strncmp( p1, p2, iLen1 ) : !PIUtil_strncmpi( p1, p2, iLen1 );
		
		if ( p1!=szBuf1 ) { PI_DELETE( [] p1 ); };
		if ( p2!=szBuf2 ) { PI_DELETE( [] p2 ); };

		if ( iResult )
			{
			if ( iLength>0 && pszBuffer )
				{ strncpy( pszBuffer, "true", 4 ); };
			return 4;
			};

		return 0;
		};

	virtual int Min()	{ return 2; };
	virtual int Max()	{ return 2; };

	/* ---
	Optimize expression. Return 1 on success, 0 on error
	--- */
	virtual int Compile( DblList &lExpressions, PIString & /* sError */ )
		{
		DblListIterator i( lExpressions );
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pEArg1 = (_Pi3Expression *)i.Current();
		i++;
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pEArg2 = (_Pi3Expression *)i.Current();

		/* --- clear the list --- */
		assert( lExpressions.Size()>=Min() );
		assert( lExpressions.Size()<=Max() );
		lExpressions.Clear();

		pArg1 = ExpressionToFunctionArgument( pEArg1 );
		pArg2 = ExpressionToFunctionArgument( pEArg2 );

		if ( !pArg1 || !pArg2 )
			{ return 0; };

		/* --- done --- */
		iOK = 1;
		return 1;
		};

	virtual int IsStatic() const
		{
		assert( pArg1 );
		assert( pArg2 );

		return 
			pArg1->eType==FunctionArgument::STRING &&
			pArg2->eType==FunctionArgument::STRING;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class NotFunctionComponent : public FunctionComponent
{
private:
	FunctionArgument *pBoolean;

public:
	NotFunctionComponent()
	:	FunctionComponent(),
		pBoolean( 0 )
		{};

	virtual ~NotFunctionComponent()
		{
 		PI_DELETE( pBoolean );		
		};

	virtual const char *GetName() const		{ return "not"; };

	/* --- */
	virtual void WriteDebugTree( int iIndent ) const
		{
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "Function:'" << GetName() << "'";
		if ( IsStatic() )
			{ cerr << "<Static>"; }
		else
			{ cerr << "<Dynamic>"; }
		assert( pBoolean );
		pBoolean->WriteDebugTree( iIndent );
		};

	/* --- */
	virtual int Write( Context &tContext, char *pszBuffer, int iLength ) const
		{
		assert( pBoolean );

		/* --- switch on the length of pBoolean, to write out --- */
		switch( pBoolean->Write( tContext, 0, 0 ) )
			{
			case -1:	return -1;
			case 0: {
				if ( iLength>0 && pszBuffer!=0 )
					{ strncpy( pszBuffer, "true", 4 ); };
				return 4;
				};
			default:	return 0;
			};
		};

	virtual int Min()	{ return 1; };
	virtual int Max()	{ return 1; };

	/* ---
	Optimize expression. Return 1 on success, 0 on error
	--- */
	virtual int Compile( DblList &lExpressions, PIString & /* sError */ ) 
		{
		DblListIterator i( lExpressions );
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pEBoolean = (_Pi3Expression *)i.Current();

		/* --- clear the list --- */
		assert( lExpressions.Size()>=Min() );
		assert( lExpressions.Size()<=Max() );
		lExpressions.Clear();

		pBoolean = ExpressionToFunctionArgument( pEBoolean );

		if ( !pBoolean )
			{ return 0; };

		/* --- done --- */
		iOK = 1;
		return 1;
		};

	virtual int IsStatic() const
		{
		assert( pBoolean );

		return pBoolean->eType==FunctionArgument::STRING;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class RegexpFunctionComponent : public FunctionComponent
{
private:
	FunctionArgument *pRegexp;
	FunctionArgument *pText;

public:
	RegexpFunctionComponent()
	:	FunctionComponent(),
		pRegexp( 0 ),
		pText( 0 )
		{};

	virtual ~RegexpFunctionComponent()
		{
 		PI_DELETE( pRegexp );		
		PI_DELETE( pText );
		};

	virtual const char *GetName() const		{ return "not"; };

	/* --- */
	virtual void WriteDebugTree( int iIndent ) const
		{
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "Function:'" << GetName() << "'";
		if ( IsStatic() )
			{ cerr << "<Static>"; }
		else
			{ cerr << "<Dynamic>"; }
		assert( pRegexp );
		pRegexp->WriteDebugTree( iIndent );
		cerr << ", ";
		assert( pText );
		pText->WriteDebugTree( iIndent );
		};

	/* --- */
	virtual int Write( Context &tContext, char *pszBuffer, int iLength ) const
		{
		assert( pRegexp );
		assert( pText );

		/* --- 1 .. Regular expression --- */
		int iRLen = pRegexp->Write( tContext, pszBuffer, iLength );
		if ( iRLen==-1 )
			{
			assert( 0 );
			return -1;
			};

		char *pR = PI_NEW( char[iRLen+1] );
		if ( !pR )
			{
			assert( 0 );
			return -1;
			};
		pR[iRLen]='\0';
		if ( iRLen>iLength )
			{ pRegexp->Write( tContext, pR, iRLen ); }
		else
			{ strncpy( pR, pszBuffer, iRLen ); };
		assert( pR[iRLen]=='\0' );

		/* --- 2 .. Text --- */
		int iTLen = pText->Write( tContext, pszBuffer, iLength );
		if ( iTLen==-1 )
			{
			PI_DELETE( [] pR );
			assert( 0 );
			return -1;
			};
		char *pT = PI_NEW( char[iTLen+1] );
		if ( !pT )
			{
			assert( 0 );
			return -1;
			};
		pT[iTLen]='\0';
		if ( iTLen>iLength )
			{ pText->Write( tContext, pT, iTLen ); }
		else
			{ strncpy( pT, pszBuffer, iTLen ); };
		assert( pT[iTLen]=='\0' );
	
		/* --- 3 .. Compare them --- */
		int iResult = HTTPUtil_regexMatch( pR, iRLen, pT, iTLen );
	
		/* --- 4 .. Deallocate memory --- */
		PI_DELETE( [] pR );
		PI_DELETE( [] pT );

		/* --- 5 .. Return result --- */
		if ( iResult )
			{
			if ( iLength && pszBuffer )
				{ strncpy( pszBuffer, "true", 4 ); };
			return 4;
			}
		else
			{ return 0; };
		};

	virtual int Min()	{ return 2; };
	virtual int Max()	{ return 2; };

	/* ---
	Optimize expression. Return 1 on success, 0 on error
	--- */
	virtual int Compile( DblList &lExpressions, PIString & /* sError */ ) 
		{
		DblListIterator i( lExpressions );
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pER = (_Pi3Expression *)i.Current();
		i++;
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pET = (_Pi3Expression *)i.Current();

		/* --- clear the list --- */
		assert( lExpressions.Size()>=Min() );
		assert( lExpressions.Size()<=Max() );
		lExpressions.Clear();

		pRegexp = ExpressionToFunctionArgument( pER );
		pText = ExpressionToFunctionArgument( pET );

		if ( !pRegexp || !pText )
			{ return 0; };

		/* --- done --- */
		iOK = 1;
		return 1;
		};

	virtual int IsStatic() const
		{
		assert( pRegexp );
		assert( pText );

		return pRegexp->eType==FunctionArgument::STRING && 
			pText->eType==FunctionArgument::STRING;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class ArgumentFunctionComponent : public FunctionComponent
{
private:
	FunctionArgument *pVariableIndex;
	FunctionArgument *pContextIndex;

	int iVariableIndex;
	int iContextIndex;

	int ToIndex( FunctionArgument *pArg, Context &tContext ) const
		{	
		if ( !pArg )
			{ return 0; };
		enum { BUF_SIZE=31 };
		char szBuf[BUF_SIZE+1];
		*szBuf='\0';
		int iLen = pArg->Write( tContext, szBuf, BUF_SIZE );
		if ( iLen==-1 )
			{ return 0; };
		if ( iLen>BUF_SIZE )
			{ iLen = BUF_SIZE; };
		szBuf[iLen]='\0';
		int iResult = atoi( szBuf );
		if ( iResult<0 )
			{
			return 0;
			};
		return iResult;
		};

public:
	ArgumentFunctionComponent()
	:	FunctionComponent(),
		pVariableIndex( 0 ),
		pContextIndex( 0 ),	
		iVariableIndex( -1 ),
		iContextIndex( -1 )
		{};

	virtual ~ArgumentFunctionComponent()
		{
 		PI_DELETE( pVariableIndex );		
		PI_DELETE( pContextIndex );
		};

	virtual const char *GetName() const		{ return "arg"; };

	/* --- */
	virtual void WriteDebugTree( int iIndent ) const
		{
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "Function:'" << GetName() << "'";
		if ( IsStatic() )
			{ cerr << "<Static>"; }
		else
			{ cerr << "<Dynamic>"; }
		if ( pVariableIndex )
			{ pVariableIndex->WriteDebugTree( iIndent ); }
		else
			{ cerr << iVariableIndex; };
		cerr << ", ";
		if ( pContextIndex )
			{ pContextIndex->WriteDebugTree( iIndent ); }
		else
			{ cerr << iContextIndex; };
		};

	/* --- */
	virtual int Write( Context &tContext, char *pszBuffer, int iLength ) const
		{
		int l_iVariableIndex = (iVariableIndex!=-1)?iVariableIndex:
				ToIndex( pVariableIndex, tContext );
		int l_iContextIndex = (iContextIndex!=-1)?iContextIndex:
				ToIndex( pContextIndex, tContext );

		/* --- */
		if ( l_iVariableIndex==-1 || l_iContextIndex==-1 ||
			( l_iVariableIndex >= Context::CONTEXT_SIZE ) )
			{
			assert( 0 );
			return -1;
			};

		/* --- find appropriate context --- */
		Context *pContext = tContext.GetContext( l_iContextIndex );

		/* --- */
		assert( l_iVariableIndex < Context::CONTEXT_SIZE );
		int iElemLength = 0;
		if ( pContext )
			{	
			iElemLength = pContext->aLength[l_iVariableIndex];
			const char *pElem = pContext->aString[l_iVariableIndex];
	
			int iLen = iLength < iElemLength ? iLength : iElemLength;
			if ( iLen>0 )
				{
				assert( pszBuffer );
				strncpy( pszBuffer, pElem, iLen );
				};
			}
		return iElemLength;
		};

	virtual int Min()	{ return 0; };
	virtual int Max()	{ return 2; };

	/* ---
	Optimize expression. Return 1 on success, 0 on error
	--- */
	virtual int Compile( DblList &lExpressions, PIString & /* sError */ )
		{
		DblListIterator i( lExpressions );
		_Pi3Expression *pEV = 0;
		_Pi3Expression *pEC = 0;

		if ( !i.BadIndex() )
			{
			pEV = (_Pi3Expression *)i.Current();
			i++;
			};
		if ( !i.BadIndex() )
			{
			pEC = (_Pi3Expression *)i.Current();
			i++;
			};

		/* --- clear the list --- */
		assert( lExpressions.Size()>=Min() );
		assert( lExpressions.Size()<=Max() );
		lExpressions.Clear();

		pVariableIndex =  ExpressionToFunctionArgument( pEV );
		pContextIndex =  ExpressionToFunctionArgument( pEC );

		/* --- make dummy PIHTTP object --- */
		PIHTTP tPIHTTP( (PIDB *)0 );
		Context tContext( 0, tPIHTTP, 0 );

		if ( pVariableIndex && pVariableIndex->eType==FunctionArgument::STRING )
			{
			iVariableIndex = ToIndex( pVariableIndex, tContext );
			PI_DELETE( pVariableIndex );
			pVariableIndex = 0;
			};

		if ( pContextIndex && pContextIndex->eType==FunctionArgument::STRING )
			{
			iContextIndex = ToIndex( pContextIndex, tContext );
			PI_DELETE( pContextIndex );
			pContextIndex = 0;
			};

		if ( !pVariableIndex && iVariableIndex==-1 )
			{ iVariableIndex = 0; };

		if ( !pContextIndex && iContextIndex==-1 )
			{ iContextIndex = 0; };

		/* --- done --- */
		iOK = 1;
		return 1;
		};

	virtual int IsStatic() const
		{
		/* ---
		Can never be static 
		--- */
		return 0;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class DBLookupFunctionComponent : public FunctionComponent
{
private:
	FunctionArgument *pDB;
	FunctionArgument *pType;
	FunctionArgument *pVariable;
	FunctionArgument *pFlags;
	FunctionArgument *pFnCallBack;
	FunctionArgument *pFirst;
	FunctionArgument *pLast;

	/* --- class local static parameters --- */
	enum TDB {	
		INVALID=0, PROGRAM, CONNECTION, HOST, REQUEST, RESPONSE } eDB;
	int iType;
	int iFlags;
	int iFirst;
	int iLast;

	/* --- internal methods --- */
	TDB ToDB( FunctionArgument *pArg, Context &tContext ) const
		{
		if ( !pArg )	
			{ return INVALID; };
		enum { BUF_SIZE=31 };
		char szBuf[BUF_SIZE+1];
		int iLen = pArg->Write( tContext, szBuf, BUF_SIZE );
		if ( iLen==-1 )
			{ return INVALID; };
		if ( iLen>BUF_SIZE )
			{ iLen = BUF_SIZE; };
		szBuf[iLen]='\0';
		const char *pszBuffer = szBuf;

		/* --- skip leading whitespace --- */
		for( ; *pszBuffer && (isspace(*pszBuffer)); pszBuffer++ );

		if ( !PIUtil_strncmpi( pszBuffer, "connection", 10 ) )
			{ return CONNECTION; }
#if 0
		/*
		** program db not implemented so the name is not 
		** recognised right now
		*/
		else if ( !PIUtil_strncmpi( pszBuffer, "program", 7 ) )
			{ return PROGRAM; }
#endif
		else if ( !PIUtil_strncmpi( pszBuffer, "host", 4 ) )
			{ return HOST; }
		else if ( !PIUtil_strncmpi( pszBuffer, "request", 7 ) )
			{ return REQUEST; }
		else if ( !PIUtil_strncmpi( pszBuffer, "response", 8 ) )
			{ return RESPONSE; }

		return INVALID;
		};

	/* --- */
	int ToType( FunctionArgument *pArg, Context &tContext ) const
		{	
		if ( !pArg )
			{
			assert( 0 );
			return -1;
			};
		enum { BUF_SIZE=31 };
		char szBuf[BUF_SIZE+1];
		int iLen = pArg->Write( tContext, szBuf, BUF_SIZE );
		if ( iLen==-1 )
			{ return -1; };
		if ( iLen>BUF_SIZE )
			{ iLen=BUF_SIZE; };
		szBuf[iLen] = '\0';
		return PIDB_pidbtypeNameToNumber( szBuf );
		};

	/* --- */
	int ToFlags( FunctionArgument *pArg, Context &tContext ) const
		{	
		if ( !pArg )
			{ return 0; };
		enum { BUF_SIZE=31 };
		char szBuf[BUF_SIZE+1];
		int iLen = pArg->Write( tContext, szBuf, BUF_SIZE );
		if ( iLen==-1 )
			{ return -1; };
		if ( iLen>BUF_SIZE )
			{ iLen=BUF_SIZE; };
		szBuf[iLen] = '\0';
		return PIDB_pidbflagsNameToNumber( szBuf );
		};

	/* --- */
	int ToIndex( FunctionArgument *pArg, Context &tContext ) const
		{	
		if ( !pArg )
			{ return 0; };
		enum { BUF_SIZE=31 };
		char szBuf[BUF_SIZE+1];
		*szBuf='\0';
		if ( pArg->Write( tContext, szBuf, BUF_SIZE )==-1 )
			{ return -1; };
		int iResult = atoi( szBuf );
		if ( iResult<0 )
			{
			return 0;
			};
		return iResult;
		};

	/* --- */
	int WriteEntity(
			Context &tContext,
			char *pszBuffer,
			int iLength,
			void *pVoid,
			int iType,
			int iVariableLength,
			const char *pszVariable,
			int iTypeLen,
			const char *pType 
			) const
		{
		/* --- buffer for numeric values --- */
		enum { NUM_BUF=63 };
		char szNumBuf[NUM_BUF+1];

		const char *pString = 0;
		int iStrLen = -1;
		switch( iType )
			{
			case PIDBTYPE_STRING:
			case PIDBTYPE_RFC822: {
				pString = pVoid ? (const char *)pVoid : "";
				iStrLen = strlen( pString );
				};
				break;

			case PIDBTYPE_OPAQUE: {
				sprintf( szNumBuf, "%ld", (unsigned long)pVoid );
				pString = szNumBuf;
				iStrLen = strlen( pString );
				};
				break;

			case PIDBTYPE_TREE: {
				pString = "**TREE**";
				iStrLen = strlen( pString );
				};
				break;

			case PIDBTYPE_USER: {
				pString = "**USER**";
				iStrLen = strlen( pString );
				};
				break;

			default:
				assert( 0 );
			};

		/* --- write out the value --- */
		if ( pFnCallBack )
			{
			Context tNewContext( &tContext, tContext.tPIHTTP, tContext.pData );
			tNewContext.aLength[0] = iStrLen;
			tNewContext.aString[0] = pString;
			tNewContext.aLength[1] = iVariableLength;
			tNewContext.aString[1] = pszVariable;
			tNewContext.aLength[2] = iTypeLen;
			tNewContext.aString[2] = pType;

			/* --- invoke pFnCallBack --- */
			return pFnCallBack->Write( tNewContext, pszBuffer, iLength );
			}
		else
			{
			if ( iLength>0 && pszBuffer )
				{ strncpy( pszBuffer, pString, iLength ); };
			return iStrLen;
			};
		};

public:
	DBLookupFunctionComponent()
	:	FunctionComponent(),
		pDB( 0 ),
		pType( 0 ),
		pVariable( 0 ),
		pFlags( 0 ),
		pFnCallBack( 0 ),
		pFirst( 0 ),
		pLast( 0 ),
		eDB( INVALID ),
		iType( -1 ),
		iFlags( -1 ),
		iFirst( -1 ),
		iLast( -1 )
		{};

	virtual ~DBLookupFunctionComponent()
		{
		PI_DELETE( pDB );
		PI_DELETE( pType );
		PI_DELETE( pVariable );
		PI_DELETE( pFlags );
		PI_DELETE( pFnCallBack );
		PI_DELETE( pFirst );
		PI_DELETE( pLast );
		};

	virtual const char *GetName() const		{ return "dblookup"; };

	/* --- */
	virtual void WriteDebugTree( int /* iIndent */ ) const
		{
#if 0
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "Function:'" << GetName() << "'";
		if ( IsStatic() )
			{ cerr << "<Static>"; }
		else
			{ cerr << "<Dynamic>"; }
		if ( pName )
			{ pName->WriteDebugTree( iIndent ); }
		else
			{ cerr << 
				(eName==CURRENT?"CURRENT":
					(eName==INDEX?"INDEX":
						(eName==SIZE?"SIZE":"INVALID"))); };
		cerr << ", ";
		if ( pContext )
			{ pContext->WriteDebugTree( iIndent ); }
		else
			{ cerr << iContextIndex; };
#endif
		};

	/* --- */
	virtual int Write( Context &tContext, char *pszBuffer, int iLength ) const
		{
		TDB l_eDB = (eDB!=INVALID)?eDB:ToDB(pDB, tContext); 
		int l_iType = (iType!=-1)?iType:ToType(pType, tContext);
		/* --- Variable --- */
		int l_iFlags = (iFlags!=-1)?iFlags:ToFlags(pFlags, tContext);
		/* --- fnValue, fnVariable, fnType --- */
		int l_iFirst = (iFirst!=-1)?iFirst:(ToIndex(pFirst, tContext));
		int l_iLast = (iLast!=-1)?iLast:(ToIndex(pLast, tContext));

		/* --- */
		if ( l_eDB == INVALID || l_iType==-1 || l_iFlags==-1 ||
			l_iFirst==-1 || l_iLast==-1 )
			{
			assert( 0 );
			return -1;
			};

		/* --- get appropriate DB --- */
		PIDB *pDB = 0;
		switch( eDB )
			{
			case PROGRAM: pDB=0; break;
			case CONNECTION: pDB=tContext.tPIHTTP.GetConnectionDB(); break;
			case HOST: pDB=tContext.tPIHTTP.GetHostDB(); break;
			case REQUEST: pDB=tContext.tPIHTTP.GetRequestDB(); break;
			case RESPONSE: pDB=tContext.tPIHTTP.GetResponseDB(); break;
			default:
				assert( 0 );
				return -1;
			};
		if ( !pDB )
			{ return 0; };

		/* --- get variable --- */
		enum { VARIABLE_BUF=1023 };
		char szBuf[VARIABLE_BUF+1];
		*szBuf='\0';
		int iVariableLength = 0;
		const char *pszVariable = 0;
		if ( pVariable )
			{
			iVariableLength = pVariable->Write( tContext, szBuf, VARIABLE_BUF );
			if ( iVariableLength==-1 )
				{ return -1; };
			if ( iVariableLength>VARIABLE_BUF )
				{
				const char aMsg[] = "** Variable name is too long **";
				if ( iLength>0 )
					{ strncpy( pszBuffer, aMsg, iLength ); };
				return sizeof( aMsg );
				};
			if ( iVariableLength > 0 )
				{ szBuf[iVariableLength] = '\0'; pszVariable = szBuf; };
			};

		/* --- setup type string --- */
		const char *pType = 0;
		int iTypeLen = -1;
		switch( l_iType )
			{
			case PIDBTYPE_STRING:	pType = "String"; iTypeLen = 6; break;
			case PIDBTYPE_RFC822:	pType = "Rfc822"; iTypeLen = 6; break;
			case PIDBTYPE_TREE:		pType = "Tree"; iTypeLen = 4; break;
			case PIDBTYPE_OPAQUE:	pType = "Opaque"; iTypeLen = 6; break;
			case PIDBTYPE_USER:		pType = "User"; iTypeLen = 4; break;
			default:
				assert( 0 );
			};

		if ( (l_iFirst==l_iLast) && (l_iFirst==0) && pszVariable &&
			!(l_iFlags & PIDBFLAG_PROPAGATEUP) )
			{
			/* --- no need for iterator --- */
			void *pVoid = PIDB_lookup( pDB, l_iType, pszVariable, l_iFlags );
			return WriteEntity( tContext, pszBuffer, iLength,
				pVoid, l_iType, iVariableLength, pszVariable,
				iTypeLen, pType );
			}
		else
			{
			PIDBIterator *pIter = PIDB_getIterator( pDB, l_iType, pszVariable,
				l_iFlags ); 
			if ( !pIter )
				{ return 0; };
			int iLengthSum = 0;
			int iRet = 0;
			for( ; PIDBIterator_atValidElement( pIter );
					PIDBIterator_next( pIter ) )
				{	
				const char *pVariable;

				void *pVoid = PIDBIterator_current( pIter, &pVariable );

				int iVarLen = iVariableLength ? iVariableLength :
					( pVariable ? strlen( pVariable ) : 0 );

				iRet = WriteEntity( tContext, pszBuffer, iLength,
					pVoid, l_iType, iVarLen, pVariable,
					iTypeLen, pType );
				if ( iRet==-1 )
					{
					assert( 0 );
					break;
					};
				iLengthSum += iRet;
				if ( iLength )
					{
					if ( iRet>=iLength )
						{ iRet = iLength; };
					iLength -= iRet;
					pszBuffer = &( pszBuffer[iRet] );
					};
				};
			PIDBIterator_delete( pIter );
			return iRet==-1 ? -1 : iLengthSum;
			};
		};

	virtual int Min()	{ return 3; };
	virtual int Max()	{ return 9; };

	/* ---
	Optimize expression. Return 1 on success, 0 on error
	--- */
	virtual int Compile( DblList &lExpressions, PIString & /* sError */ )
		{
		DblListIterator i( lExpressions );
		if ( i.BadIndex() ) { return 0; }
		_Pi3Expression *pEDB = (_Pi3Expression *)i.Current(); i++;
		if ( i.BadIndex() ) { return 0; }
		_Pi3Expression *pEType = (_Pi3Expression *)i.Current(); i++;
		if ( i.BadIndex() ) { return 0; }
		_Pi3Expression *pEVariable = (_Pi3Expression *)i.Current(); i++;
		_Pi3Expression *pEFlags = 0;
		_Pi3Expression *pEFnValue = 0;
		_Pi3Expression *pEFnVariable = 0;
		_Pi3Expression *pEFnType = 0;
		_Pi3Expression *pEFirst = 0;
		_Pi3Expression *pELast = 0;
		if ( !i.BadIndex() ) { pEFlags = (_Pi3Expression *)i.Current(); i++; };
		if ( !i.BadIndex() ) { pEFnValue = (_Pi3Expression *)i.Current(); i++; };
		if ( !i.BadIndex() ) { pEFnVariable = (_Pi3Expression *)i.Current(); i++; };
		if ( !i.BadIndex() ) { pEFnType = (_Pi3Expression *)i.Current(); i++; };
		if ( !i.BadIndex() ) { pEFirst = (_Pi3Expression *)i.Current(); i++; };
		if ( !i.BadIndex() ) { pELast = (_Pi3Expression *)i.Current(); i++; };

		/* --- clear the list --- */
		assert( lExpressions.Size()>=Min() );
		assert( lExpressions.Size()<=Max() );
		lExpressions.Clear();

		pDB = ExpressionToFunctionArgument( pEDB );
		pType = ExpressionToFunctionArgument( pEType );
		pVariable = ExpressionToFunctionArgument( pEVariable );
		pFlags = ExpressionToFunctionArgument( pEFlags );
		pFnCallBack = ExpressionToFunctionArgument( pEFnValue );
		pFirst = ExpressionToFunctionArgument( pEFirst );
		pLast = ExpressionToFunctionArgument( pELast );

		/* --- make dummy PIHTTP object --- */
		PIHTTP tPIHTTP( (PIDB *)0 );
		Context tContext( 0, tPIHTTP, 0 );

		/* --- optimize --- */
		if ( pDB && pDB->eType==FunctionArgument::STRING )
			{ eDB = ToDB( pDB, tContext ); PI_DELETE( pDB ); pDB = 0; };

		if ( pType && pType->eType==FunctionArgument::STRING )
			{iType = ToType( pType, tContext );PI_DELETE(pType);pType = 0;};

		/* --- check if the variable is empty --- */
		if ( pVariable && pVariable->eType==FunctionArgument::STRING &&
				pVariable->Write( tContext, (char *)0, 0 )==0 )
			{ PI_DELETE(pVariable);pVariable = 0;};

		if ( pFlags && pFlags->eType==FunctionArgument::STRING )
			{iFlags = ToFlags( pFlags, tContext );PI_DELETE(pFlags);pFlags = 0;};

		if ( pFirst && pFirst->eType==FunctionArgument::STRING )
			{iFirst=ToIndex( pFirst, tContext );PI_DELETE(pFirst);pFirst = 0;};

		if ( pLast && pLast->eType==FunctionArgument::STRING )
			{iLast = ToIndex( pLast, tContext );PI_DELETE(pLast);pLast = 0;};

		if ( pDB && pDB->eType==FunctionArgument::STRING )
			{ eDB = ToDB( pDB, tContext ); PI_DELETE( pDB ); pDB = 0; };

		/* --- validation mandatory parameters --- */
		if (	( !pDB && eDB == INVALID ) || 
				( !pType && iType==-1 ) )
			{ return 0; };

		/* --- done --- */
		iOK = 1;
		return 1;
		};

	virtual int IsStatic() const
		{
		/* ---
		Can never be static 
		--- */
		return 0;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class DBReplaceFunctionComponent : public FunctionComponent
{
private:
	FunctionArgument *pDB;
	FunctionArgument *pType;
	FunctionArgument *pVariable;
	FunctionArgument *pNewValue;
	FunctionArgument *pFlags;
	PIString sNewValue;

	/* --- class local static parameters --- */
	enum TDB {	
		INVALID=0, PROGRAM, CONNECTION, HOST, REQUEST, RESPONSE } eDB;
	int iType;
	int iFlags;

	/* --- internal methods --- */
	TDB ToDB( FunctionArgument *pArg, Context &tContext ) const
		{
		if ( !pArg )	
			{ return INVALID; };
		enum { BUF_SIZE=31 };
		char szBuf[BUF_SIZE+1];
		int iLen = pArg->Write( tContext, szBuf, BUF_SIZE );
		if ( iLen==-1 )
			{ return INVALID; };
		if ( iLen>BUF_SIZE )
			{ iLen = BUF_SIZE; };
		szBuf[iLen]='\0';
		const char *pszBuffer = szBuf;

		/* --- skip leading whitespace --- */
		for( ; *pszBuffer && (isspace(*pszBuffer)); pszBuffer++ );

		if ( !PIUtil_strncmpi( pszBuffer, "program", 7 ) )
			{ return PROGRAM; }
		else if ( !PIUtil_strncmpi( pszBuffer, "connection", 10 ) )
			{ return CONNECTION; }
		else if ( !PIUtil_strncmpi( pszBuffer, "host", 4 ) )
			{ return HOST; }
		else if ( !PIUtil_strncmpi( pszBuffer, "request", 7 ) )
			{ return REQUEST; }
		else if ( !PIUtil_strncmpi( pszBuffer, "response", 8 ) )
			{ return RESPONSE; }

		return INVALID;
		};

	/* --- */
	int ToType( FunctionArgument *pArg, Context &tContext ) const
		{	
		if ( !pArg )
			{
			assert( 0 );
			return -1;
			};
		enum { BUF_SIZE=31 };
		char szBuf[BUF_SIZE+1];
		int iLen = pArg->Write( tContext, szBuf, BUF_SIZE );
		if ( iLen==-1 )
			{ return -1; };
		if ( iLen>BUF_SIZE )
			{ iLen=BUF_SIZE; };
		szBuf[iLen] = '\0';
		return PIDB_pidbtypeNameToNumber( szBuf );
		};

	/* --- */
	int ToFlags( FunctionArgument *pArg, Context &tContext ) const
		{	
		if ( !pArg )
			{ return 0; };
		enum { BUF_SIZE=31 };
		char szBuf[BUF_SIZE+1];
		int iLen = pArg->Write( tContext, szBuf, BUF_SIZE );
		if ( iLen==-1 )
			{ return -1; };
		if ( iLen>BUF_SIZE )
			{ iLen=BUF_SIZE; };
		szBuf[iLen] = '\0';
		return PIDB_pidbflagsNameToNumber( szBuf );
		};

	void ToString( FunctionArgument *pArg, Context &tContext,
		PIString &sValue ) const
		{
		int iLen = pArg->Write( tContext, 0, 0 );
		char *pBuf = PI_NEW( char[iLen+1] );	
		pArg->Write( tContext, pBuf, iLen );	
		pBuf[iLen] = '\0';
		sValue = pBuf;
		PI_DELETE( [] pBuf );
		};

public:
	DBReplaceFunctionComponent()
	:	FunctionComponent(),
		pDB( 0 ),
		pType( 0 ),
		pVariable( 0 ),
		pNewValue( 0 ),
		pFlags( 0 ),
		eDB( INVALID ),
		iType( -1 ),
		iFlags( -1 )
		{};

	virtual ~DBReplaceFunctionComponent()
		{
		PI_DELETE( pDB );
		PI_DELETE( pType );
		PI_DELETE( pVariable );
		PI_DELETE( pNewValue );
		PI_DELETE( pFlags );
		};

	virtual const char *GetName() const		{ return "dbreplace"; };

	/* --- */
	virtual void WriteDebugTree( int /* iIndent */ ) const
		{
#if 0
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "Function:'" << GetName() << "'";
		if ( IsStatic() )
			{ cerr << "<Static>"; }
		else
			{ cerr << "<Dynamic>"; }
		if ( pName )
			{ pName->WriteDebugTree( iIndent ); }
		else
			{ cerr << 
				(eName==CURRENT?"CURRENT":
					(eName==INDEX?"INDEX":
						(eName==SIZE?"SIZE":"INVALID"))); };
		cerr << ", ";
		if ( pContext )
			{ pContext->WriteDebugTree( iIndent ); }
		else
			{ cerr << iContextIndex; };
#endif
		};

	/* --- */
	virtual int Write( Context &tContext, char *pszBuffer, int iLength ) const
		{
		TDB l_eDB = (eDB!=INVALID)?eDB:ToDB(pDB, tContext); 
		int l_iType = (iType!=-1)?iType:ToType(pType, tContext);
		int l_iFlags = (iFlags!=-1)?iFlags:ToFlags(pFlags, tContext);

		/* --- */
		if ( l_eDB == INVALID || l_iType==-1 || l_iFlags==-1 )
			{ return 0; };

		/* --- get appropriate DB --- */
		PIDB *pDB = 0;
		switch( eDB )
			{
			case PROGRAM: pDB=0; break;
			case CONNECTION: pDB=tContext.tPIHTTP.GetConnectionDB(); break;
			case HOST: pDB=tContext.tPIHTTP.GetHostDB(); break;
			case REQUEST: pDB=tContext.tPIHTTP.GetRequestDB(); break;
			case RESPONSE: pDB=tContext.tPIHTTP.GetResponseDB(); break;
			default:
				assert( 0 );
			};
		if ( !pDB )
			{ return 0; };

		const char *pValue = 0;
		PIString sTmp;
		if ( pNewValue )
			{
			ToString( pNewValue, tContext, sTmp );
			pValue = sTmp;
			}
		else
			{ pValue = sNewValue; };

		const char *pTheVariable = 0;
		PIString sTmp2;
		if ( pVariable )
			{
			ToString( pVariable, tContext, sTmp2 );
			pTheVariable = sTmp2;
			}
		else
			{ pTheVariable = 0; };

		PIDB_replace( pDB, l_iType, pTheVariable, (void *)pValue, l_iFlags );

		if ( iLength )
			{
			assert( pszBuffer );
			int iLen = (iLength>=4) ? 4 : iLength; 
			strncpy( pszBuffer, "true", iLen );
			};
		return 4;
		};

	virtual int Min()	{ return 3; };
	virtual int Max()	{ return 9; };

	/* ---
	Optimize expression. Return 1 on success, 0 on error
	--- */
	virtual int Compile( DblList &lExpressions, PIString & /* sError */ )
		{
		DblListIterator i( lExpressions );
		if ( i.BadIndex() ) { return 0; }
		_Pi3Expression *pEDB = (_Pi3Expression *)i.Current(); i++;
		if ( i.BadIndex() ) { return 0; }
		_Pi3Expression *pEType = (_Pi3Expression *)i.Current(); i++;
		if ( i.BadIndex() ) { return 0; }
		_Pi3Expression *pEVariable = (_Pi3Expression *)i.Current(); i++;
		if ( i.BadIndex() ) { return 0; }
		_Pi3Expression *pENewValue = (_Pi3Expression *)i.Current(); i++;
		_Pi3Expression *pEFlags = 0;
		if ( !i.BadIndex() ) { pEFlags = (_Pi3Expression *)i.Current(); i++; };

		/* --- clear the list --- */
		assert( lExpressions.Size()>=Min() );
		assert( lExpressions.Size()<=Max() );
		lExpressions.Clear();

		pDB = ExpressionToFunctionArgument( pEDB );
		pType = ExpressionToFunctionArgument( pEType );
		pVariable = ExpressionToFunctionArgument( pEVariable );
		pNewValue = ExpressionToFunctionArgument( pENewValue );
		pFlags = ExpressionToFunctionArgument( pEFlags );

		/* --- make dummy PIHTTP object --- */
		PIHTTP tPIHTTP( (PIDB *)0 );
		Context tContext( 0, tPIHTTP, 0 );

		/* --- optimize --- */
		if ( pDB && pDB->eType==FunctionArgument::STRING )
			{ eDB = ToDB( pDB, tContext ); PI_DELETE( pDB ); pDB = 0; };

		if ( pType && pType->eType==FunctionArgument::STRING )
			{iType = ToType( pType, tContext );PI_DELETE(pType);pType = 0;};

		/* --- check if the variable is empty --- */
		if ( pVariable && pVariable->eType==FunctionArgument::STRING &&
				pVariable->Write( tContext, (char *)0, 0 )==0 )
			{ PI_DELETE(pVariable);pVariable = 0;};

		if ( pFlags && pFlags->eType==FunctionArgument::STRING )
			{iFlags = ToFlags( pFlags, tContext );PI_DELETE(pFlags);pFlags = 0;};

		if ( pDB && pDB->eType==FunctionArgument::STRING )
			{ eDB = ToDB( pDB, tContext ); PI_DELETE( pDB ); pDB = 0; };

		if ( pNewValue && pNewValue->eType==FunctionArgument::STRING )
			{
			ToString( pNewValue, tContext, sNewValue);
			PI_DELETE( pNewValue );
			pNewValue = 0;
			};

		/* --- validation mandatory parameters --- */
		if (	( !pDB && eDB == INVALID ) || 
				( !pType && iType==-1 ) )
			{ return 0; };

		/* --- done --- */
		iOK = 1;
		return 1;
		};

	virtual int IsStatic() const
		{
		/* ---
		Can never be static 
		--- */
		return 0;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class AbbrevFunctionComponent : public FunctionComponent
{
private:
	FunctionArgument *pField;
	FunctionArgument *pFieldLen;
	FunctionArgument *pAppend;
	int iFieldLen;

	int ToNumber( Context &tContext, FunctionArgument *pArg ) const
		{
		enum { BUF_SIZE=63 };
		char szBuf[BUF_SIZE+1];
		int iLen = pArg->Write( tContext, szBuf, BUF_SIZE );
		if ( iLen==-1 )
			{ return -1; };
		if ( iLen==0 )
			{ return 0; };
		int i = 0;
		if ( iLen>BUF_SIZE )
			{
			char *pTmp = PI_NEW( char[iLen+1] );
			pTmp[iLen]='\0';
			i = atoi( pTmp );
			PI_DELETE( [] pTmp );
			}
		else 
			{
			szBuf[iLen]='\0';
			i = atoi( szBuf );
			};
		return i;
		};

public:
	AbbrevFunctionComponent()
	:	FunctionComponent(),
		pField( 0 ),
		pFieldLen( 0 ),
		pAppend( 0 ),
		iFieldLen( -1 )
		{};

	virtual ~AbbrevFunctionComponent()
		{
 		PI_DELETE( pField );		
		PI_DELETE( pFieldLen );
		PI_DELETE( pAppend );
		};

	virtual const char *GetName() const		{ return "abbrev"; };

	/* --- */
	virtual void WriteDebugTree( int iIndent ) const
		{
		for(int i=0; i<iIndent; i++ )
			{ cerr << INDENT_PATTERN; };
		cerr << "Function:'" << GetName() << "'";
		if ( IsStatic() )
			{ cerr << "<Static>"; }
		else
			{ cerr << "<Dynamic>"; }
		assert( pField );
		pField->WriteDebugTree( iIndent );
		cerr << ", ";
		assert( pFieldLen || iFieldLen!=-1 );
		if ( pFieldLen )
			{ pFieldLen->WriteDebugTree( iIndent ); }
		else
			{ cerr << iFieldLen; };
		cerr << ", ";
		assert( pAppend );
		pAppend->WriteDebugTree( iIndent );
		};

	/* --- */
	virtual int Write( Context &tContext, char *pszBuffer, int iLength ) const
		{
		assert( pFieldLen || iFieldLen!=-1 );
		assert( pField );
		assert( pAppend );

		/* --- 1 .. FieldLen --- */
		int l_iFieldLen = iFieldLen!=-1 ? iFieldLen : ToNumber( tContext,
			pFieldLen );
		if ( l_iFieldLen==-1 )
			{ return -1; };
		if ( l_iFieldLen==0 )
			{ return 0; };

		/* --- 2 .. Field --- */
		int iTmpLen = l_iFieldLen<iLength ? l_iFieldLen : iLength;
		int iFLen = pField->Write( tContext, pszBuffer, iTmpLen );
		if ( iFLen==-1 )
			{ return -1; };
		if ( iFLen<=l_iFieldLen )
			{ return iFLen; };

		/* --- 3 .. Append --- */
		int iALen = pAppend->Write( tContext, 0, 0 );
		if ( iALen==-1 )
			{ return -1; };

		/* --- adjust iALen to be length of append sequence --- */
		if ( iALen>l_iFieldLen )
			{ iALen = l_iFieldLen; };

		/* --- adjust pszBuffer and iLength up to where append is --- */
		pszBuffer = &( pszBuffer[l_iFieldLen-iALen] );
		iLength -= ( l_iFieldLen-iALen );

		if ( iLength>0 )
			{ pAppend->Write( tContext, pszBuffer, iLength ); };

		return l_iFieldLen;
		};

	virtual int Min()	{ return 3; };
	virtual int Max()	{ return 3; };

	/* ---
	Optimize expression. Return 1 on success, 0 on error
	--- */
	virtual int Compile( DblList &lExpressions, PIString & /* sError */ ) 
		{
		DblListIterator i( lExpressions );
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pEF = (_Pi3Expression *)i.Current();
		i++;
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pEFL = (_Pi3Expression *)i.Current();
		i++;
		if ( i.BadIndex() ) { return 0; };
		_Pi3Expression *pEA = (_Pi3Expression *)i.Current();

		/* --- make dummy PIHTTP object --- */
		PIHTTP tPIHTTP( (PIDB *)0 );
		Context tContext( 0, tPIHTTP, 0 );

		/* --- clear the list --- */
		assert( lExpressions.Size()>=Min() );
		assert( lExpressions.Size()<=Max() );
		lExpressions.Clear();

		pField = ExpressionToFunctionArgument( pEF );
		pFieldLen = ExpressionToFunctionArgument( pEFL );
		pAppend = ExpressionToFunctionArgument( pEA );

		if ( !pField || !pFieldLen || !pAppend )
			{ return 0; };

		/* --- see can field be converted to number --- */
		if ( pFieldLen->eType==FunctionArgument::STRING )
			{
			iFieldLen = ToNumber( tContext, pFieldLen );
			PI_DELETE( pFieldLen );
			pFieldLen = 0;
			};

		/* --- done --- */
		iOK = 1;
		return 1;
		};

	virtual int IsStatic() const
		{
		assert( pField );
		assert( pAppend );

		return
			pField->eType==FunctionArgument::STRING && 
			iFieldLen!=-1 &&
			pAppend->eType==FunctionArgument::STRING;
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
FunctionComponent *FunctionComponent::MakeFunctionComponent(
	PIString &sFunctionName, TokenSource &tSource,
	FnPi3Write *pFunctions, PIString &sError,
	int &iNestedMessage /* --- error message is from nested expression --- */ )
{
	sError = PIString::Empty();

	PIOStrStream os;
	DblList lExpressions;

	FunctionComponent *pComp = 0;
	int iMin = 0;
	int iMax = 0;
	int i = 0;

	if ( sFunctionName=="trunc" )
		{ pComp = PI_NEW( TruncFunctionComponent() ); }
	else if ( sFunctionName=="align" )
		{ pComp = PI_NEW( AlignFunctionComponent() ); }
	else if ( sFunctionName=="if" )
		{ pComp = PI_NEW( IfFunctionComponent() ); }
	else if ( sFunctionName=="not" )
		{ pComp = PI_NEW( NotFunctionComponent() ); }
	else if ( sFunctionName=="regexp" )
		{ pComp = PI_NEW( RegexpFunctionComponent() ); }
	else if ( sFunctionName=="arg" || sFunctionName=="_" )
		{ pComp = PI_NEW( ArgumentFunctionComponent() ); }
	else if ( sFunctionName=="dblookup" )
		{ pComp = PI_NEW( DBLookupFunctionComponent() ); }
	else if ( sFunctionName=="dbreplace" )
		{ pComp = PI_NEW( DBReplaceFunctionComponent() ); }
	else if ( sFunctionName=="abbrev" )
		{ pComp = PI_NEW( AbbrevFunctionComponent() ); }
	else if ( sFunctionName=="cmp" )
		{ pComp = PI_NEW( CmpFunctionComponent( 1 ) ); }
	else if ( sFunctionName=="cmpi" )
		{ pComp = PI_NEW( CmpFunctionComponent( 0 ) ); }
	else if ( sFunctionName=="and" )
		{ pComp = PI_NEW( AndFunctionComponent() ); }
	else if ( sFunctionName=="or" )
		{ pComp = PI_NEW( OrFunctionComponent() ); }

	/* --- its an error if the function was not found --- */
	if ( !pComp )
		{
		os << "Unknown function '" << sFunctionName << "'." << ends;
		sError = os.str();
		goto error;
		};

	/* --- read in the functions parameters --- */
	for(;;)
		{
		Token *pNext = tSource.GetNextToken( sError );
		_Pi3Expression *pExpr = 0;

		if ( !pNext )
			{
			if ( sError!=PIString::Empty() )
				{ sError = "End of expression reached missing ')'."; };
			goto error;
			};

		if ( pNext->eType==Token::CLOSE_PAREN )
			{
			PI_DELETE( pNext );
			break;
			}
		else
			{
			/* --- push back so the expression will absort it --- */
			tSource.PushBack( pNext );
			};

		pExpr = PI_NEW( _Pi3Expression( tSource, pFunctions, sError ) );

		if ( !pExpr || !pExpr->IsOK() )
			{
			PI_DELETE( pExpr );
			iNestedMessage = 1;
			goto error;
			};

		lExpressions.Append( (DblList::type) pExpr );
		};

	/* --- check function parameters --- */
	iMin = pComp->Min();
	iMax = pComp->Max();
	i = lExpressions.Size();
	if ( i<iMin || i>iMax )
		{
		os << "Function '" << sFunctionName << "' ";
		if ( iMin==iMax )
			{
			os << "takes exactly " << iMin << " parameters.";
			}
		else
			{
			os << "takes between " << iMin << " and " << iMax
				<< " parameters.";
			};

		os << ends;
		sError = os.str();
		goto error;
		};

	/* --- derived class processing --- */
	if ( !pComp->Compile( lExpressions, sError ) )
		{ goto error; };

	/* --- done --- */
	{
	for( DblListIterator k( lExpressions ); !k.BadIndex(); k++ ) 
		{ PI_DELETE( (_Pi3Expression *)k.Current() ); };
	};

	return pComp;

error:
	PI_DELETE( pComp );
	for( DblListIterator j( lExpressions ); !j.BadIndex(); j++ ) 
		{ PI_DELETE( (_Pi3Expression *)j.Current() ); };
	if ( sError==PIString::Empty() )
		{
		sError = "Error verifying function parameters for function '";
		sError.Concatenate( sFunctionName );
		sError.Concatenate( "'." );
		};

	return 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Return non-zero on success, 0 on error.
\*____________________________________________________________________________*/
int Internal_ParseExpression(
		TokenSource &tSource,
		FnPi3Write *pFunctions,
		PIString &sError,
		DblList &lComponents,
		int iIsSubExpression )
{
	enum {
		NORMAL,
		DOLLAR_SEEN,
		PERCENT_SEEN, 
		AMPERSAND_SEEN,
		FUNCTION_DECL
		} eState = NORMAL;

	sError = PIString::Empty();
	PIString sBuffer;

#define EMIT_STRING {				\
		lComponents.Append( (DblList::type)new StringComponent( sBuffer ) ); \
		sBuffer = ""; }

#define ADD_CHAR(c) { sBuffer.Concatenate( c ); }

	enum { TERM_LINE_LEN=80 };
#define SYNTAX_ERROR_RETURN( msg ) { \
		int i=tSource.GetPosition(); \
		const char *pExpr = tSource.GetRawString(); \
		if ( i>60 ) { pExpr = &( pExpr[i-60] ); i = 60; };\
		PIOStrStream os; \
		os << msg << endl; \
		PIString sTrunc( pExpr, 75 ); \
		os << '[' << sTrunc << ']' << endl; \
		for(int j=0; j<i; j++) { os << ' '; }; \
		os << '^' << endl; \
		for(int k=4; k<i; k++) { os << ' '; }; \
		os << "near here" << endl; \
		os << ends; \
		sError = os.str(); \
		return 0; \
		}

	/* --- cross token persistent values --- */
	PIString sFunctionName;		/* name of next function */

	for(;;)
		{
		Token *pNext = tSource.GetNextToken( sError );
		Token::Type t = pNext ? pNext->eType : Token::OTHER;
		if ( !pNext || ( iIsSubExpression &&
			( t==Token::CLOSE_PAREN || t==Token::COMMA ) )
			)
			{
			if ( pNext )
				{
				/* --- push back token if its a ')' --- */
				if ( pNext->eType==Token::CLOSE_PAREN )
					{ tSource.PushBack( pNext ); }
				/* --- delete it if its a comma --- */
				else if ( pNext->eType==Token::COMMA )
					{ PI_DELETE( pNext ); };
				}
			else
				{
				if ( sError!=PIString::Empty() )
					{
					PIString sMyError; /* ( sError ); */
					SYNTAX_ERROR_RETURN( sMyError );
					};
				};

			switch( eState )
				{
				case NORMAL:	EMIT_STRING; return 1;
				case DOLLAR_SEEN:	ADD_CHAR( '%' ); return 1;
				case PERCENT_SEEN:	ADD_CHAR( '$' ); return 1;
				case AMPERSAND_SEEN:	ADD_CHAR( '&' ); return 1;
				case FUNCTION_DECL: {
						SYNTAX_ERROR_RETURN( "'(' expected:" );
						};
				default: assert( 0 ); return 0;
				};
			};

		/* --- make AutoDelete object to make sure Token is deleted --- */
		AutoDelete<Token> tA( pNext );
		(void)tA;
		switch( eState )
			{
			case NORMAL:
				switch( t )
					{
					case Token::DOLLAR:
						EMIT_STRING; eState = DOLLAR_SEEN; break;
					case Token::PERCENT:
						EMIT_STRING; eState = PERCENT_SEEN; break;
					case Token::AMPERSAND:
						EMIT_STRING; eState = AMPERSAND_SEEN; break;

					case Token::OPEN_PAREN:	ADD_CHAR( '(' ); break;
					case Token::CLOSE_PAREN: ADD_CHAR( ')' ); break;
					case Token::COMMA: ADD_CHAR( ',' ); break;
					case Token::IDENTIFIER: 
						sBuffer.Concatenate( pNext->sStr ); break;
					case Token::LITERAL: 
						sBuffer.Concatenate( pNext->sStr ); break;

					default:
						assert( t==Token::OTHER );
						ADD_CHAR( pNext->cChar );
						break;
					};
				break;

			case DOLLAR_SEEN: {
				eState = NORMAL;
				if ( t==Token::DOLLAR )
					{
					ADD_CHAR( '$' );
					break;
					};
				if ( t!=Token::IDENTIFIER )
					{
					PIString sDescription;
					pNext->GetDescription( sDescription );
					SYNTAX_ERROR_RETURN( "Identifier expected, not '" <<
							sDescription << "':" );
					return 0;
					};
				const char *pTmp = pNext->sStr;
				char c = *pTmp;
				ComponentBase *pComp = PI_NEW( ShortcutComponent( c ) );
				if ( !pComp || !pComp->IsOK() )
					{
					PIString sDescription;
					pNext->GetDescription( sDescription );
					PI_DELETE( pComp );
					SYNTAX_ERROR_RETURN("Unknown shortcut '"<< (c) <<"':" );
					return 0;
					}
				lComponents.Append( (DblList::type)pComp );
				pTmp = pNext->sStr;
				pNext->sStr = &( pTmp[1] );
				tSource.PushBack( PI_NEW( Token( *pNext ) ) );
				break;
				};

			case PERCENT_SEEN: {
				eState = NORMAL;
				if ( t==Token::PERCENT )
					{
					ADD_CHAR( '%' );
					break;
					};
				if ( t!=Token::IDENTIFIER )
					{
					PIString sDescription;
					pNext->GetDescription( sDescription );
					SYNTAX_ERROR_RETURN( "Identifier expected, not '" <<
							sDescription << "':" );
					return 0;
					};
				const char *pTmp = pNext->sStr;
				char c = *pTmp;
				ComponentBase *pComp = PI_NEW( ParameterComponent(
					pFunctions ? (pFunctions[(c)]) : 0 ) );
				if ( !pComp || !pComp->IsOK() )
					{
					PIString sDescription;
					pNext->GetDescription( sDescription );
					PI_DELETE( pComp );
					SYNTAX_ERROR_RETURN("Unknown parameter '"<< (c) <<"':" );
					return 0;
					}
				lComponents.Append( (DblList::type)pComp );
				pTmp = pNext->sStr;
				pNext->sStr = &( pTmp[1] );
				tSource.PushBack( PI_NEW( Token( *pNext ) ) );
				break;
				};
				
			case AMPERSAND_SEEN:
				if ( t==Token::AMPERSAND )
					{
					ADD_CHAR( '&' );
					break;
					}
				else if ( t!=Token::IDENTIFIER )
					{
					PIString sDescription;
					pNext->GetDescription( sDescription );
					SYNTAX_ERROR_RETURN( "Identifier expected, not '" <<
							sDescription << "':" );
					return 0;
					};
				sBuffer = pNext->sStr;
				eState = FUNCTION_DECL;
				break;

			case FUNCTION_DECL: {
				if ( t!=Token::OPEN_PAREN )
					{
					PIString sDescription;
					pNext->GetDescription( sDescription );
					SYNTAX_ERROR_RETURN( "'(' expected, not '" <<
							sDescription << "':" );
					return 0;
					};
				eState = NORMAL;
				PIString sMyError;
				int iNestedMessage = 0;
				ComponentBase *pComp =
					FunctionComponent::MakeFunctionComponent( sBuffer,
						tSource, pFunctions, sMyError, iNestedMessage ); 
				if ( !pComp || !pComp->IsOK() )
					{
					PI_DELETE( pComp );
					if ( iNestedMessage )
						{
						/* --- line position already given --- */
						sError = sMyError;
						return 0;
						};

					SYNTAX_ERROR_RETURN( sMyError );
					assert( 0 );	/* --- how'd we get here ? --- */
					return 0;
					}
				lComponents.Append( (DblList::type)pComp );
				sBuffer = PIString::Empty();
				break;
				};

			default:
				assert( 0 ); return 0;
			};
		};

	assert( 0 );
	return 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
_Pi3Expression::_Pi3Expression()
:	pExpr( 0 ),
	iOK( 1 )
{
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Internal Version, invoked recursively when functions load thier 
	parameter expressions.
\*____________________________________________________________________________*/
_Pi3Expression::_Pi3Expression( TokenSource &tSource,
	FnPi3Write *pFunctions, PIString &sError )
:	pExpr( 0 ),
	iOK( 0 )
{
	DblList lRawComponents;
	if ( !Internal_ParseExpression( tSource, pFunctions, sError,
		lRawComponents, 1 ) )
		{ return; };

	PIHTTP tPIHTTP( (PIDB *)0 );
	Context tContext( 0, tPIHTTP, 0 );
	pExpr = PI_NEW( ExpressionComponent( tContext, lRawComponents, sError ) );
	assert( pExpr );
	if ( !pExpr->IsOK() )
		{
		PI_DELETE( pExpr );
		pExpr = 0;
		return;
		};

	iOK = 1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
_Pi3Expression::_Pi3Expression( const char *pTheExpression,
	FnPi3Write *pFunctions, PIString &sError )
:	pExpr( 0 ),
	iOK( 0 )
{
	assert( pTheExpression );
	TokenSource tSource( pTheExpression );
	DblList lRawComponents;
	if ( !Internal_ParseExpression( tSource, pFunctions, sError,
		lRawComponents, 0 ) )
		{
		for( DblListIterator i( lRawComponents ); !i.BadIndex(); i++ )
			{ PI_DELETE( (ComponentBase *)i.Current() ); };
		return;
		};

	PIHTTP tPIHTTP( (PIDB *)0 );
	Context tContext( 0, tPIHTTP, 0 );
	pExpr = PI_NEW( ExpressionComponent( tContext, lRawComponents, sError ) );
	assert( pExpr );

	if ( !pExpr->IsOK() )
		{
		PI_DELETE( pExpr );
		pExpr = 0;
		return;
		};

	iOK = 1;
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
_Pi3Expression::~_Pi3Expression()
{
	PI_DELETE( pExpr );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Returns -1 on error, else number of characters written.
\*____________________________________________________________________________*/
int _Pi3Expression::Write(
		PIHTTP &tPIHTTP,
		void *pData,
		char *pszBuffer,
		int iMaxLen ) const
{
	Context tContext( 0, tPIHTTP, pData );
	int iRet = Write( tContext, pszBuffer, iMaxLen );
	assert( iRet!=-1 );
	return iRet;
}

