/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Server/MThread.cpp,v $
 * $Date: 2002/01/26 19:50:42 $
 *
 Description:
	IO server class, derived from IOServerBase class.

	Multithreaded dispatch of IO requests.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <assert.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>

#include "IOSrvBse.h"
#include "Pi2API.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_MAXTHREADS		"MaxThreads"
#define KEY_CONF_EXITAFTER		"ExitAfter"

#define CONFIG_ERR(this, msg)   \
	{ PILog_addMessage( PIObject_getDB(this), \
	PIObject_getConfigurationDB(this), PILOG_ERROR, (msg) ); }
#define CONFIG_WARN(this, msg)   \
	{ PILog_addMessage( PIObject_getDB(this), \
	PIObject_getConfigurationDB(this), PILOG_WARNING, (msg) ); }
/*
#define D { cerr << __FILE__ << ": " << __LINE__ << endl; };
*/
#define D

/* ---
	MUSTBETRUE - alway execute the statement, additionally assert it
	during development 
--- */
#if !defined(NDEBUG)
#	define MUSTBETRUE(x)	(assert(x))
#else
#	define MUSTBETRUE(x)	(x)
#endif

/* --
	Determine whether we use thread-per-request or thread-reuse model
--- */
#if 1
#	define THREAD_PER_REQUEST 1	/* thread-reuse not ready for prime time! */
#endif

/*____________________________________________________________________________*\
 *
 Description:
	Documentation.
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this server
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	MultiThreadedIOServer

Description:
	This object performs multithreaded dispatch of IO requests. On
	execution it waits for a new IO request and dispatches it to a
	thread for processing. It then returns to dispatch another request.
	<P>
	This dispatching object initializes an IO object for accepting IO 
	requests, and a logic object for handling those requests.

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>IOObject
<TD>+
<TD>A Pi3 IO object name
<TD>A Pi3 object
<TD>IOObject "ServerTCPIPObject"

<TR>
<TD>LogicObject
<TD>+
<TD>A Pi3 Logic object name
<TD>A Pi3 object
<TD>LogicObject "HandleRequestObject"

<TR>
<TD>MaxThreads
<TD>+
<TD>A non-negative
<TD>Number of threads to use
<TD>MaxThreads 10

<TR>
<TD>ExitAfter
<TD>-1
<TD>An integer
<TD>Number of request to exit after
<TD>ExitAfter 1000

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>
	IOObject
</H5>
Specifies the IO object used to accept new IO requests. The IO object is
created during construction. During execution of this object PIObject_copy()
is invoked to create a copy IO object for each new request. When processing
of the request is completed, the new IO object is destroyed and the server
waits for another IO request by again calling PIObject_copy() on the main
IO object.
The IO object can be an object which provides an IO service (such as TCP/IP,
named Pipe or shared memory IO) or an IO filter which performs translation
on such an IO object (such as encryption or monitoring).

<H5>
	LogicObject
</H5>
Specifies the logic object used to handle IO requests. The object is 
created during construction and its <I>execute</I> method invoked
(using PILogic_execute()) for each IO request accepted. The IO object
for the request is passed as an argument to this logic object.

<H5>
	MaxThreads
</H5>
Specifies the maximum number of threads to use to concurrently handle
IO requests. While no more threads are availablable to new IO requests
will be accepted.
This number should be between 1 and 1000.

<H5>
	ExitAfter
</H5>
Specifying a positive number for this value will cause the this server
object to call exit() after that number of requests. Threads with outstanding
requests will be allow to complete before exit() is invoked.
<P>
This should only be used in the context of a process group where a monitor
process fork()s child processes to replace ones that die. In that
scenario it is good for preventing longterm memory or resource leaks.
<P>
This directive has no effect on non-UNIX systems.

Returns:
	PIAPI_COMPLETED on success.
	PIAPI_ERROR and PIAPI_ABORT respectively for generic and severe
	error conditions.

Example:
	<PRE>
	&lt;Object&gt;
		Name MultiThreadedIOServer
		Class MultiThreadedIOServerClass
		IOObject ServerTCPIPIO
		LogicObject HandleHTTP
	&lt;/Object&gt;
	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class MultiThreadedIOServer : public IOServerBase
{
private:
	class ThreadRecord
		{
		public:
		PIThread *pThread;
#if !THREAD_PER_REQUEST
		PISync *pSema;
#endif
		MultiThreadedIOServer *pServer;
		PIObject *pIOObject;
		ThreadRecord( PIThread *pTheThread )
			:	pThread( pTheThread ),
#if !THREAD_PER_REQUEST
				pSema( PIPlatform_allocLocalSemaphore( 0, 2 ) ),
#endif
				pServer( 0 ),
				pIOObject( 0 )
			{
			};

		~ThreadRecord()
			{
#if !THREAD_PER_REQUEST
			PISync_delete( pSema );
#endif
			};
		};
	PISync *pFreeThreadsMutex;		/* guard access to free threads array */
	PISync *pFreeThreadsSemaphore;	/* # threads is a resource, count usage */
	int iNumBusyThreads;			/* count the number of busy threads */
	ThreadRecord **ppAllThreads;	/* array of all thread records */
	ThreadRecord **ppFreeThreads;	/* array of records of free threads */
	int iMaxThreads;
	int iRequestsToLive;			/* exit after this number of requests */

	/* ---
	Automatically invoked from IOServerBase::Accept()
	This overrides the default simple DoDispatchConnection() in
	IOServerBase (which just calls DispatchConnection() directly, to
	allocate a thread and make the thread run DispatchConnection()
	--- */
	virtual int DoDispatchConnection( PIObject *pIOObject )
		{
		/* ---
		Select a thread to run, access to the free threads list is
		protected by the pFreeThreadsMutex.
		--- */
		MUSTBETRUE( !PISync_lock( pFreeThreadsMutex ) );
		ThreadRecord *pThreadRecord = 0;
		for( int i=0; i<iMaxThreads; i++ )
			{
			if ( ppFreeThreads[i] )
				{
				pThreadRecord = ppFreeThreads[i];
				ppFreeThreads[i] = 0;
				break;
				};
			};
		/* ---	
		Another thread is officially busy
		--- */
		iNumBusyThreads++;
		MUSTBETRUE( !PISync_unlock( pFreeThreadsMutex ) );
		PIThread *pThread = pThreadRecord->pThread;
		assert( pThreadRecord->pThread );
		assert( pThread );
		
#if THREAD_PER_REQUEST
		/* ---
		Wait for this thread to fully terminate, this prevents
		a race condition where a thread adds itself to the free
		list but gets preempted before terminating
		--- */
		PIThread_waitForThread( pThread );
		pThreadRecord->pIOObject = pIOObject;
		pThreadRecord->pServer = this;
		if ( PIThread_begin(
			pThread,
			(PIThreadFn)MultiThreadedIOServer::StaticDispatchConnection,
			(unsigned long)pThreadRecord,
			PITHREAD_PRIORITY_HIGH, 0 ) ) 
			{
			MUSTBETRUE( !PISync_unlock( pFreeThreadsSemaphore ) );
			MUSTBETRUE( !PIObject_delete( pIOObject, 0, 0 ) );
	
			/* --- consider this a critical error --- */
			return PIAPI_ABORT;
			};
#else
		pThreadRecord->pIOObject = pIOObject;
		MUSTBETRUE( !PISync_unlock( pThreadRecord->pSema ) );
#endif
		return PIAPI_COMPLETED;
		};

	/* ---
	Called by StaticDispatchConnection() this function	
	marks the current thread as free so it may be reused
	--- */
	void FreeThread( ThreadRecord *pFree )
		{

		/* --- add this thread back to the free list --- */
		assert( pFreeThreadsMutex );
		MUSTBETRUE( !PISync_lock( pFreeThreadsMutex ) );
		int i=0; for(; i<iMaxThreads; i++ )
			{
			if ( !ppFreeThreads[i] )
				{ ppFreeThreads[i] = pFree; break; };
			};
		/* ---
		Another thread is officially free
		--- */
		iNumBusyThreads--;
		MUSTBETRUE( !PISync_unlock( pFreeThreadsMutex ) );
		MUSTBETRUE( !PISync_unlock( pFreeThreadsSemaphore ) );
		assert( i<iMaxThreads );
		};

	/* ---
	Called by the threads begin function this runs in a thread
	created for this connection.
	--- */
	inline static int StaticDispatchConnection( ThreadRecord *pThreadRecord )
		{
		assert( pThreadRecord );
		if ( !pThreadRecord )
			{
			assert( 0 );
			return PIAPI_ERROR;
			};
		for(;;)
			{
			/* --- wait on the mutex --- */
#if !THREAD_PER_REQUEST
			if ( PISync_lock( pThreadRecord->pSema ) )
				{
				MUSTBETRUE( 0 );
				};
#endif

			MultiThreadedIOServer *pServer = pThreadRecord->pServer;
			if ( !pServer )
				{
				/* --- message to exit the thread --- */
				break;
				};

			PIObject *pIOObject = pThreadRecord->pIOObject;
 			pServer->DispatchConnection( pIOObject );
			pServer->FreeThread( pThreadRecord );

#if THREAD_PER_REQUEST
			break;
#endif
			};
		return PIAPI_COMPLETED;
		}

public:
	MultiThreadedIOServer( const PIObject *pTheObject )
	:	IOServerBase( pTheObject ),
		pFreeThreadsMutex( PIPlatform_allocLocalMutex() ),
		pFreeThreadsSemaphore( 0 ),
		iNumBusyThreads( 0 ),
		ppAllThreads( 0 ),
		ppFreeThreads( 0 ),
		iMaxThreads( 0 ),
		iRequestsToLive( -1 )
		{
		if ( !bOK ) { return;	/* Base class initialization failed */ };
		bOK = 0;				/* until initialization is complete */
		if ( !pFreeThreadsMutex )
			{
			CONFIG_ERR( pObject, "MultiThreadedIOServer: Could not \
allocate mutex.");
			return;
			};
		PIConfig *pConfigDB = PIObject_getConfigurationDB( pObject );

		/* ---
		Get exit after count
		--- */
		const char *pExitAfter = PIConfig_lookupValue( pConfigDB,
			KEY_CONF_EXITAFTER, 0, 0 );
		if ( pExitAfter )
			{
			iRequestsToLive = atoi( pExitAfter );
			iRequestsToLive = iRequestsToLive > 0 ? iRequestsToLive : -1 ;
			}; 

		const char *pMaxThreads = PIConfig_lookupValue( pConfigDB,
			KEY_CONF_MAXTHREADS, 0, 0 );
		if ( !pMaxThreads )
			{
			CONFIG_ERR( pObject, \
"MultiThreadedIOServer: Missing definition for 'MaxThreads'");
			return;
			};
		iMaxThreads = atoi( pMaxThreads );
		if ( iMaxThreads<1 || iMaxThreads>1000 )
			{
			CONFIG_WARN( pObject, "MultiThreadedIOServer: Invalid number for 'MaxThreads', \
defaulting to 25");
			iMaxThreads = 25; 
			return;
			};
		pFreeThreadsSemaphore = PIPlatform_allocLocalSemaphore( iMaxThreads,
			iMaxThreads );
		if ( !pFreeThreadsSemaphore )
			{
			CONFIG_ERR( pObject, "MultiThreadedIOServer: Could not allocate semaphore.");
			return;
			};
		ppAllThreads = new ThreadRecord *[iMaxThreads];
		ppFreeThreads = new ThreadRecord *[iMaxThreads];
		if ( !ppFreeThreads || !ppAllThreads )
			{ assert( 0 ); return; };
		memset( ppAllThreads, 0, sizeof( ThreadRecord * ) * iMaxThreads );
		memset( ppFreeThreads, 0, sizeof( ThreadRecord * ) * iMaxThreads );
		int iGood = 1;
		for( int i=0; iGood && i<iMaxThreads; i++ )
			{
			PIThread *pNew = PIThread_new( 0, 0 );
			ThreadRecord *pRecord = new ThreadRecord( pNew );
			ppFreeThreads[i] = pRecord;
			ppAllThreads[i] = pRecord;
#if !THREAD_PER_REQUEST
			pRecord->pServer = this;
			if ( PIThread_begin(
				pNew,
				(PIThreadFn)MultiThreadedIOServer::StaticDispatchConnection,
				(unsigned long)pRecord,
				PITHREAD_PRIORITY_HIGH, 0 ) ) 
				{
				assert( 0 );
				CONFIG_ERR( pObject, \
"MultiThreadedIOServer: Failed to prime thread, PIThread_begin() failed.");
				iGood = 0;
				};
#endif
			};

		if ( ppAllThreads && ppFreeThreads && iGood )
			{ bOK = 1; };
		};

	virtual ~MultiThreadedIOServer()
		{
		for( int i=0; i<iMaxThreads; i++ )
			{
			PIThread *pThread = ppAllThreads[i]->pThread;
#if !THREAD_PER_REQUEST
			/* --- signal thread to die --- */
			ppAllThreads[i]->pServer = 0;
			PISync_unlock( ppAllThreads[i]->pSema );
			MUSTBETRUE( !PIThread_waitForThread( pThread ) );
#endif
			MUSTBETRUE( !PIThread_delete( pThread ) );
			delete ppAllThreads[i];
			};
		delete [] ppAllThreads;
		delete [] ppFreeThreads;

		MUSTBETRUE( 
			!pFreeThreadsMutex || !PISync_delete( pFreeThreadsMutex ) );
		MUSTBETRUE(
			!pFreeThreadsSemaphore || !PISync_delete( pFreeThreadsSemaphore ) );
		};

	virtual int Accept()
		{
		/* ---
		Check is it time for this server to exit
		--- */
#if POSIX
		if ( iRequestsToLive )
			{ 	
			if ( iRequestsToLive>0 )
				{ iRequestsToLive--; };
			}
		else
			{
			/* ---
			Time to go, gracefully exit.
			--- */
			while( iNumBusyThreads>0 )
				{
				MUSTBETRUE( !PISync_lock( pFreeThreadsSemaphore ) );
				};
#if 1
			/*
			** Why even bother?
			*/
			/* ---
			Allow all threads to die
			--- */
			for( int i=0; i<iMaxThreads; i++ )
				{
				/* --- signal thread to die --- */
				ppAllThreads[i]->pServer = 0;
#if !THREAD_PER_REQUEST
				PISync_unlock( ppAllThreads[i]->pSema );
#endif		
				};
#endif
			::exit( 0 );
			};
#endif

		/* ---
		Before attempting to accept a connection, make sure there
		is a thread available to handle it
		--- */
		MUSTBETRUE( !PISync_lock( pFreeThreadsSemaphore ) );
		
		/* ---
		Accept a new child IO object
		--- */
		PIObject *pChildIO = PIObject_copy( GetPrototypeIOObject(), 0, 0 );
		if ( !pChildIO )
			{
			/* --- consider this a critical error --- */
			CONFIG_ERR( pObject, "MultiThreadedIOServer: PIObject_copy() \
failed for IO Object");
			MUSTBETRUE( !PISync_unlock( pFreeThreadsSemaphore ) );
			return PIAPI_ABORT;
			};

		return DoDispatchConnection( pChildIO );
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int MultiThreadedIOServer_onClassLoad( PIClass_LoadAction eLoad, void * )
{
	switch( eLoad )
		{
		case STARTUP:
		default:;
		};

	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int MultiThreadedIOServer_constructor( PIObject *pObj, int, const char *[] )
{
	MultiThreadedIOServer *pServer = new MultiThreadedIOServer( pObj );
	if (!( pServer && pServer->IsOK() ))
		{
		delete pServer;
		return PIAPI_ERROR;
		};
	if ( PIObject_setUserData( pObj, pServer ) )
		{
		CONFIG_ERR( pObj, "MultiThreadedIOServer_constructor: PIObject_setUserData() failed" );
		return PIAPI_ERROR;
		};
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int MultiThreadedIOServer_copyConstructor( PIObject *, int,
	const char *[] )
{
	/* --- forbid copy construction --- */
	return PIAPI_ERROR;
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name MultiThreadedIOServerClass
		Type LogicExtension
		Library Server
		OnClassLoad MultiThreadedIOServer_onClassLoad
		Constructor MultiThreadedIOServer_constructor
		CopyConstructor MultiThreadedIOServer_copyConstructor
		Destructor IOServerBase_destructor
		Execute IOServerBase_execute
	</Class>

	<Object>
		Name MultiThreadedIOServer
		Class MultiThreadedIOServerClass
	</Object>

/*___+++CNF_END+++___*/
#endif

